"""
Given a sequence, generate a PLM embedding augmented with
in-silico mutagenesis and CDR isolation.
"""

from __future__ import annotations
import argparse
from abmap.abmap_augment import augment_from_fasta

from typing import Callable, NamedTuple


class AugmentArguments(NamedTuple):
    cmd: str
    device: int
    outdir: str
    seqs: str
    chain_type: str
    plm: str
    num_mutations: int
    separators: bool
    cdr_masks: bool
    func: Callable[[AugmentArguments], None]


def add_args(parser):
    """
    Create parser for command line utility.
    :meta private:
    """
    parser.add_argument(
        "--seqs", help="Sequences to be embedded", required=True
    )
    parser.add_argument(
        "-o", "--outdir", help="h5 file to write results", required=True
    )
    parser.add_argument(
        "-d", "--device", type=int, default=-1, help="Compute device to use"
    )
    parser.add_argument(
        "--chain-type", dest="chain_type",
        help="Chain type of the fasta sequences (H or L)", required=True
    )
    parser.add_argument(
        "--plm", help="which foundational PLM to use for embedding", required=True
    )
    parser.add_argument(
        "--num-mutations", type=int, default=100, dest='num_mutations',
        help="Number of mutations for in-silico mutagenesis. 0 means no mutations."
    )
    parser.add_argument(
        "--separators", action="store_true", dest='separators',
        help="Place separators (0 pad) between the CDRs"
    )
    parser.add_argument(
        "--cdr-masks", action="store_true", dest='cdr_masks',
        help="concatenate 4 extra dimensions for CDR position masks."
    )
    return parser


def main(args):
    """
    Run embedding from arguments.
    :meta private:
    """
    inPath = args.seqs
    outPath = args.outdir
    device = args.device
    num_mutations = args.num_mutations
    chain_type = args.chain_type
    embed_type = args.plm
    seperators = args.separators
    cdr_masks = args.cdr_masks
    augment_from_fasta(inPath, outPath, chain_type, embed_type,
                       num_mutations, seperators, cdr_masks, 
                       device, verbose=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    add_args(parser)
    main(parser.parse_args())


import os
import sys
import argparse
from tqdm import tqdm
import glob
import time
import numpy as np
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt
import pickle
import pandas as pd
import torch
import h5py

from abmap.utils import get_boolean_mask, find_sequence, parse, log
from abmap.mutate import generate_mutation
from abmap.plm_embed import embed_sequence, reload_models_to_device


class ProteinEmbedding:

    def __init__(self, sequence, chain_type, embed_device=None, embed_model=None, dev=0, fold=0):
        self.seq = sequence
        self.chain_type = chain_type
        self.embedding = None
        self.cdr_mask = None
        self.cdr_embedding = None
        self.embed_device = embed_device
        self.embed_model = embed_model
        self.dev = dev
        self.fold = fold

        assert self.chain_type in ['H', 'L']

    def embed_seq(self, embed_type = "beplerberger"):
        self.embedding = embed_sequence(self.seq, embed_type = embed_type, 
                                        embed_device=self.embed_device, 
                                        embed_model=self.embed_model)

    def create_cdr_mask(self, scheme='chothia', buffer_region = False):
        self.cdr_mask = get_boolean_mask(self.seq, self.chain_type, scheme, buffer_region, self.dev, self.fold)

    def mutate_seq(self, p=0.5, mut_type='cat2'):
        if self.cdr_mask != None:
            return generate_mutation(self.seq, self.cdr_mask, p, mut_type)
        else:
            print("please first create the CDR masks!")
            raise ValueError

    def create_embeds_noaug(self, num_muts = 50, mut_type = 'cat1', embed_type = 'beplerberger'):
        """
        given a sequence of length n, amd embedding type with dimension d,
        return a average mutated embeeding matrix of same shape (n x d)
        """

        mutation_embeds = []

        # print("performing {} mutagenesis of {} mutations...".format(mut_type, num_muts))
        for i in range(num_muts):
            mut_seq = self.mutate_seq(p = 0.1, mut_type = mut_type)
            mut_embed = embed_sequence(mut_seq, embed_type, 
                                        embed_device=self.embed_device, 
                                        embed_model=self.embed_model)
            mutation_embeds.append(mut_embed)

        mutation_embeds = torch.stack(mutation_embeds)
        output = torch.mean(mutation_embeds, dim=0)

        return output

    def create_kmut_matrix(self, num_muts = 50, mut_type = 'cat2', embed_type = 'beplerberger'):
        """
        given a sequence of length n, and embedding type with dimension d,
        return a mutated CDR embedding difference matrix of size (num_muts x n x d)
        """

        mutation_embeds = []

        # print("performing {} mutagenesis of {} mutations...".format(mut_type, num_muts))
        for i in range(num_muts):
            mut_seq = self.mutate_seq(mut_type = mut_type)
            mut_embed = embed_sequence(mut_seq, embed_type,
                                        embed_device=self.embed_device, 
                                        embed_model=self.embed_model)
            mutation_embeds.append(mut_embed)

        mutation_embeds = torch.stack(mutation_embeds)

        return mutation_embeds

    def create_embedding_whole(self, kmut_matrix, verbose=False):
        diff_matrix = self.embedding - kmut_matrix
        avg_kmut = torch.mean(diff_matrix, dim=0)

        cdr_label_all = (self.cdr_mask > 0).float()
        cdr_label_1 = (self.cdr_mask == 1).float()
        cdr_label_2 = (self.cdr_mask == 2).float()
        cdr_label_3 = (self.cdr_mask == 3).float()
        cdr_label = torch.stack([cdr_label_1, cdr_label_2, cdr_label_3, cdr_label_all], dim=-1)

        if verbose:
            return torch.cat((self.embedding.cpu(), avg_kmut.cpu(), cdr_label), dim=-1)
        else:
            out = torch.cat((avg_kmut.cpu(), cdr_label), dim=-1)
            assert out.shape[-1] == 6169
            return out

    def create_cdr_embedding(self, kmut_matrix, sep = False, mask = False):
        """
        given a (k x n x d) k-mutations matrix, compute the CDR embedding 
        by averaging over the k mutation matrices and then concatenating the CDR regions 
        of the sequence --> (n* x d)
        """

        diff_matrix = self.embedding - kmut_matrix
        avg_kmut = torch.mean(diff_matrix, dim=0)

        assert avg_kmut.shape[0] == len(self.cdr_mask)

        cdr1, cdr2, cdr3 = [], [], []
        for i, cdr in enumerate(self.cdr_mask):
            if cdr == 1:
                cdr1.append(i)
            elif cdr == 2:
                cdr2.append(i)
            elif cdr == 3:
                cdr3.append(i)

        out = None
        cdr_idxs = cdr1 + cdr2 + cdr3
        if sep:
            emb_dim = avg_kmut.shape[-1]
            out = torch.cat((avg_kmut[cdr1, :].cuda(), torch.zeros(1, emb_dim).cuda(),
                              avg_kmut[cdr2, :].cuda(), torch.zeros(1, emb_dim).cuda(),
                              avg_kmut[cdr3, :].cuda()), dim=0)
        else:
            out = avg_kmut[cdr_idxs, :]

        if mask:
            if sep:
                print("Cannot use mask augmentions when there are separator tokens.")
                raise ValueError
            cdr_mask_loc = self.cdr_mask[cdr_idxs]
            cdr_label_all = (cdr_mask_loc > 0).float()
            cdr_label_1 = (cdr_mask_loc == 1).float()
            cdr_label_2 = (cdr_mask_loc == 2).float()
            cdr_label_3 = (cdr_mask_loc == 3).float()
            cdr_label = torch.stack([cdr_label_1, cdr_label_2, cdr_label_3, cdr_label_all], dim=-1)

            out = torch.cat((out.cpu(), cdr_label), dim=-1)

        self.cdr_embedding = out

        return out

    def create_cdr_embedding_nomut(self, embed_type, sep = False, mask=False):
        """
        splice the CDR regions of the embedding together with no mutation adjustments
        if sep is True, tensor of 0's are inserted as pad's between each CDR region
        """
        self.embed_seq(embed_type = embed_type)

        self.create_cdr_mask()

        result_matr = self.embedding.detach().clone()

        assert result_matr.shape[0] == len(self.cdr_mask)

        cdr1, cdr2, cdr3 = [], [], []
        for i, cdr in enumerate(self.cdr_mask):
            if cdr == 1:
                cdr1.append(i)
            elif cdr == 2:
                cdr2.append(i)
            elif cdr == 3:
                cdr3.append(i)

        if sep:
            emb_dim = result_matr.shape[-1]
            out_embed = torch.cat((result_matr[cdr1, :].cuda(), torch.zeros(1, emb_dim).cuda(),
                              result_matr[cdr2, :].cuda(), torch.zeros(1, emb_dim).cuda(),
                              result_matr[cdr3, :].cuda()), dim=0)
        else:
            cdr_idxs = cdr1 + cdr2 + cdr3
            out_embed = result_matr[cdr_idxs, :]

        out_embed = out_embed.detach().cpu()

        if mask:
            if sep:
                print("Cannot use mask augmentions when there are separator tokens.")
                raise ValueError
            cdr_mask_loc = self.cdr_mask[cdr_idxs]
            cdr_label_all = (cdr_mask_loc > 0).float()
            cdr_label_1 = (cdr_mask_loc == 1).float()
            cdr_label_2 = (cdr_mask_loc == 2).float()
            cdr_label_3 = (cdr_mask_loc == 3).float()
            cdr_label = torch.stack([cdr_label_1, cdr_label_2, cdr_label_3, cdr_label_all], dim=-1)

            out_embed = torch.cat((out_embed.cpu(), cdr_label), dim=-1)

        return out_embed


    def create_cdr_specific_embedding(self, embed_type, k=100, separator = False, mask = True):
        """
        Create a CDR-specific embedding directly from sequence
        embed_type : embed with which general purpose PLM? (i.e. beplerberger)
        k : number of mutant sequences you'd wish to generate
        separator : separate the CDR regions with 0-tensor tokens? (default: False)
        mask : Appending a mask that indicates which CDR a residue belongs to (1, 2, 3) (default: True)
        """

        self.embed_seq(embed_type = embed_type)

        self.create_cdr_mask()

        kmut_matr_h = self.create_kmut_matrix(num_muts=k, embed_type=embed_type)
        cdr_embed = self.create_cdr_embedding(kmut_matr_h, sep = separator, mask = mask)

        return cdr_embed
            

def augment_from_fasta(fastaPath, outputPath, chain_type, embed_type,
                       num_mutations, separators, cdr_masks,
                       device=0, verbose=False):
    """
    Embed sequences with existing PLMs and augment with mutagenesis & CDR isolation
    
    :param fastaPath: Input sequence file (``.fasta`` format)
    :type fastaPath: str
    :param outputPath: Output embedding file (``.h5`` format)
    :type outputPath: str
    :param device: Compute device to use for embeddings [default: 0]
    :type device: int
    :param verbose: Print embedding progress
    :type verbose: bool
    """

    reload_models_to_device(device, embed_type)

    use_cuda = (device >= 0) and torch.cuda.is_available()
    if use_cuda:
        torch.cuda.set_device(device)
        if verbose:
            log(f"# Using CUDA device {device} - {torch.cuda.get_device_name(device)}")
    else:
        if verbose:
            log("# Using CPU")

    if verbose:
        log("# Loading Model...")

    names, seqs = parse(fastaPath)

    # check if outputPath is an existing directory:
    if not os.path.isdir(outputPath):
        os.mkdir(outputPath)


    log("# Storing to {}...".format(outputPath))
    # with torch.no_grad(), h5py.File(outputPath, "a") as h5fi:
    try:
        for (name, seq) in tqdm(zip(names, seqs), total=len(names)):
            fname = os.path.join(outputPath, f'{name}.p')
            if not os.path.isfile(fname):
            # if name not in h5fi:
                prot = ProteinEmbedding(seq, chain_type, embed_device=f'cuda:{device}', dev=device)
                if num_mutations > 0:
                    z = prot.create_cdr_specific_embedding(embed_type, num_mutations,
                                                                        separators, cdr_masks)
                else:
                    z = prot.create_cdr_embedding_nomut(embed_type, separators, cdr_masks)

                with open(fname, 'wb') as f:
                    pickle.dump(z.cpu(), f)

    except KeyboardInterrupt:
        sys.exit(1)




import torch
import pandas as pd
import numpy as np
from tqdm import tqdm
import os, sys
import h5py
import pickle
import random

from torch.nn.utils.rnn import pad_sequence

from abmap.abmap_augment import ProteinEmbedding
from abmap.model import AbMAPAttn
from abmap.plm_embed import reload_models_to_device

def generate_embeddings(args):
    reload_models_to_device(args.device_num)

    device = torch.device("cuda:{}".format(args.device_num) if torch.cuda.is_available() else "cpu")

    data_dir = "/data/cb/rsingh/work/antibody/ci_data/raw/thera-sabdab/TheraSAbDab_SeqStruc_OnlineDownload.csv"

    df = pd.read_csv(data_dir)
    # focus only on Whole mAb
    df = df.loc[df['Format'] == 'Whole mAb']
    df.reset_index(drop=True, inplace=True)

    chains_feats = {'H':[], 'L':[]}
    phaseI, est_status, discontinuous = [], [], []
    thera_ids = []
    for idx, row in tqdm(df.iterrows(), total = df.shape[0]):

        thera_id = row['Therapeutic']
        seq_h, seq_l = row['Heavy Sequence'], row['Light Sequence']

        phaseI.append("Phase-I" in row["Highest_Clin_Trial (Oct '21)"])
        est_status.append("Active" in row["Est. Status"])
        discontinuous.append("Discontinued" in row["Est. Status"])

        for seq, chain_type in [(seq_h, 'H'), (seq_l, 'L')]:
            prot = ProteinEmbedding(seq, chain_type, embed_device="cuda:{}".format(args.device_num), dev=0)

            try:
                prot.create_cdr_mask()
            except:
                chains_feats[chain_type].append(None)
                continue

            k=50
            prot.embed_seq('beplerberger')
            kmut_matr_h = prot.create_kmut_matrix(num_muts=k, embed_type='beplerberger')
            cdr_embed = prot.create_cdr_embedding(kmut_matr_h, sep = False, mask = True)

            chains_feats[chain_type].append(cdr_embed)
            if chain_type == 'H':
                thera_ids.append(thera_id)

    # only keep features where both heavy and light chain features are present:
    H_final, L_final = [], []
    phaseI_final, est_status_final, discontinuous_final = [], [], []
    for i in range(len(chains_feats['H'])):
        if chains_feats['H'][i] is not None and chains_feats['L'][i] is not None:
            H_final.append(chains_feats['H'][i])
            L_final.append(chains_feats['L'][i])
            phaseI_final.append(phaseI[i])
            est_status_final.append(est_status[i])
            discontinuous_final.append(discontinuous[i])


    save_dir = '/data/cb/rsingh/work/antibody/ci_data/processed/thera-sabdab/cdrembed_maskaug4_n{}.h5'.format(len(chains_feats['H']))
    if os.path.exists(save_dir):
        os.system('rm -f {}'.format(save_dir))

    # pack the dataset into a single tensor
    cdrembeds_H = pad_sequence(H_final, batch_first=True)
    cdrembeds_L = pad_sequence(L_final, batch_first=True)
    phaseI_final = np.array(phaseI_final)
    est_status_final = np.array(est_status_final)
    discontinuous_final = np.array(discontinuous_final)
    assert len(cdrembeds_H) == len(cdrembeds_L) == len(phaseI_final) == len(est_status_final)

    print("Dataset Shapes, H: {}, L: {}".format(cdrembeds_H.shape, cdrembeds_L.shape))

    # SAVE IT
    with h5py.File(save_dir, "w") as f:
        dset_H = f.create_dataset('H', cdrembeds_H.shape)
        dset_H[:,:,:] = cdrembeds_H

        dset_L = f.create_dataset('L', cdrembeds_L.shape)
        dset_L[:,:,:] = cdrembeds_L

        dset_p = f.create_dataset('Phase-I', phaseI_final.shape)
        dset_p[:] = phaseI_final

        dset_a = f.create_dataset('Active', est_status_final.shape)
        dset_a[:] = est_status_final

        dset_d = f.create_dataset('Discontinued', discontinuous_final.shape)
        dset_d[:] = discontinuous_final

    with open('/data/cb/rsingh/work/antibody/ci_data/processed/thera-sabdab/thera_ids_n{}.p'.format(len(discontinuous_final)), 'wb') as f:
        pickle.dump(thera_ids, f)




def generate_features_from_embeddings(args):
    thera_dir = "/data/cb/rsingh/work/antibody/ci_data/processed/thera-sabdab/cdrembed_maskaug4_n547.h5"
    out_path = "/data/cb/rsingh/work/antibody/ci_data/processed/thera-sabdab/thera_features"
    chain_type = args.chain_type

    with h5py.File(thera_dir, 'r') as f:
        thera_embeds = np.array(f[chain_type])
        print(thera_embeds.shape)
    with open("/data/cb/rsingh/work/antibody/ci_data/processed/thera-sabdab/thera_ids_n547.p", 'rb') as p:
        thera_ids_list = pickle.load(p)

    print(len(thera_ids_list), len(thera_embeds))
    assert len(thera_ids_list) == len(thera_embeds)

    for i in tqdm(range(len(thera_embeds))):
        feat_np = thera_embeds[i:i+1, :, :]
        idx = unpad_embedding(feat_np)
        if idx == 0:
            embedding = torch.FloatTensor(feat_np).to(device)
        else:
            embedding = torch.FloatTensor(feat_np[:,:idx,:]).to(device)

        with torch.no_grad():
            thera_feat, _ = pretrained(embedding, embedding, None, None, task=0, task_specific=True)
            thera_feat = thera_feat.detach().cpu()
        assert thera_feat.shape[-1] == 512
        with open(os.path.join(out_path, "{}_struc_{}.p".format(thera_ids_list[i], chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(thera_feat), p)
        
        with torch.no_grad():
            thera_feat, _ = pretrained(embedding, embedding, None, None, task=1, task_specific=True)
            thera_feat = thera_feat.detach().cpu()
        assert thera_feat.shape[-1] == 512
        with open(os.path.join(out_path, "{}_func_{}.p".format(thera_ids_list[i], chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(thera_feat), p)
        
        with torch.no_grad():
            thera_feat, _ = pretrained(embedding, embedding, None, None, task=None, return2=True)
            thera_feat = thera_feat.detach().cpu()
        assert thera_feat.shape[-1] == 256
        with open(os.path.join(out_path, "{}_interm_{}.p".format(thera_ids_list[i], chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(thera_feat), p)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", help="which code path to run. see main(..) for details", type=str, choices = ["1_generate_embeddings", "2_generate_features"])
    parser.add_argument("--device_num", help="GPU device for computation.", type=int, default=0)
    parser.add_argument("--extra", help="put this as the LAST option and arbitrary space-separated key=val pairs after that", type=str, nargs='*')

    args = parser.parse_args()
    assert args.mode is not None
    
    extra_args = dict([a.split("=") for a in args.extra]) if args.extra else {}
    args.extra = extra_args
    
    pd.set_option('use_inf_as_na', True)

    
    if args.mode == '1_generate_embeddings':
        generate_embeddings(args)
    elif args.mode == '2_generate_features':
        generate_features(args)
    else:
        assert False




import os
import torch
from torch import nn
import torch.nn.functional as F

try:
    from positional_encodings.torch_encodings import PositionalEncoding1D
except:
    from positional_encodings import PositionalEncoding1D

def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad is True)



class AbMAPAttn(nn.Module):
    def __init__(self, embed_dim = 6165, mid_dim1 = 2048, mid_dim2 = 512, mid_dim3 = 128,
                  proj_dim = 50, hidden_dim = 50, num_enc_layers = 1, num_heads=5):
        '''
            
        '''
        assert (proj_dim + 4)%num_heads == 0, "AbMAPAttn parameter proj_dim + 4 must be divisible by num_heads. i.e. (proj_dim + 4)/num_heads is an int"
        super(AbMAPAttn, self).__init__()
        self.embed_dim = embed_dim
        self.activation = nn.LeakyReLU()
        
        if embed_dim >= 2048:
            self.project = nn.Sequential(
                                nn.Linear(embed_dim, mid_dim2),
                                nn.LayerNorm(mid_dim2),
                                self.activation,
                                nn.Linear(mid_dim2, mid_dim3),
                                nn.LayerNorm(mid_dim3),
                                self.activation,
                                nn.Linear(mid_dim3, proj_dim),
                                nn.LayerNorm(proj_dim),
                                self.activation,
                                nn.Dropout(p=0.5),
                              )

        else:
            self.project = nn.Sequential(
                                nn.Linear(embed_dim, mid_dim2),
                                nn.LayerNorm(mid_dim2),
                                self.activation,
                                nn.Linear(mid_dim2, proj_dim),
                                nn.LayerNorm(proj_dim),
                                self.activation,
                                nn.Dropout(p=0.5),
                              )


        self.cossim = nn.CosineSimilarity(dim=-1)

        # last layer initialization
        self.transform2_s = nn.Linear(1, 1)
        self.transform2_s.weight = torch.nn.Parameter(torch.ones(1, 1))
        self.transform2_s.bias = torch.nn.Parameter(torch.zeros(1))

        self.transform2_f = nn.Linear(1, 1)
        self.transform2_f.weight = torch.nn.Parameter(torch.ones(1, 1))
        self.transform2_f.bias = torch.nn.Parameter(torch.zeros(1))


        self.enc_layer_s = nn.TransformerEncoderLayer(d_model=proj_dim+4, nhead=num_heads, batch_first=True,
                                                      dim_feedforward=proj_dim, dropout=0.5)
        self.enc_layer_f = nn.TransformerEncoderLayer(d_model=proj_dim+4, nhead=num_heads, batch_first=True,
                                                      dim_feedforward=proj_dim, dropout=0.5)

        self.attention_s = nn.TransformerEncoder(encoder_layer=self.enc_layer_s, num_layers=num_enc_layers)
        self.attention_f = nn.TransformerEncoder(encoder_layer=self.enc_layer_f, num_layers=num_enc_layers)

        self.posenc = PositionalEncoding1D(proj_dim)


    def attention(self, x, x_mask, task):
        if task == 0:
            mha = self.attention_s
        if task == 1:
            mha = self.attention_f

        if x_mask is None:
            output = mha(x)
        else:
            output = mha(x, src_key_padding_mask=x_mask)

        return output

    def mean_over_seq(self, x, x_mask):
        if x.shape[-1] > 6000:
            x = x[:,:,-2204:-4]
        else:
            x = x[:,:,:-4]

        x = torch.sum(x, dim=1)
        x = torch.div(x, torch.unsqueeze(x_mask.shape[-1] - torch.sum(x_mask, dim=-1), dim=-1))
        
        return x


    def embed(self, x, x_mask=None, task='structure', embed_type='variable'):
        assert embed_type in ['variable', 'fixed']
        assert task in ['structure', 'function']
        ### Get variablee length embedding
        # Project to a lower dimension through several linear + LayerNorm layers
        if x.shape[-1] > 6000:
            x, x_pos = x[:,:,-4-self.embed_dim:-4], x[:,:,-4:]
        else:
            x, x_pos = x[:,:,:-4], x[:,:,-4:]
        x = self.project(x)

        # ADD POSITIONAL ENCODING HERE:
        x = torch.add(x, self.posenc(x))
        x = torch.cat([x, x_pos], dim=-1)

        # Attention block
        att_task = 0 if task == 'structure' else 1
        x = self.attention(x, x_mask, att_task)

        # Output variable-length embedding as an option
        if embed_type == 'variable': return x

        ### Get Fixed-length embedding
        # method 1: mean of the features over the seq_len:
        if x_mask is None:
            x_ = torch.mean(x, dim=1)
        else:
            x_ = torch.sum(x, dim=1)
            x_ = torch.div(x_, torch.unsqueeze(x_mask.shape[-1] - torch.sum(x_mask, dim=-1), dim=-1))

        # method 2: log-sum-exp of the output features from attention layers
        x = torch.log(torch.sum(torch.exp(x), dim=1))

        # ---------------------------------------
        # concatenate mean and logsumexp
        x = F.normalize(x)
        x_ = F.normalize(x_)
        x = torch.cat((x_, x), dim=-1)
        # ---------------------------------------

        # Output fixed-length embedding here. This is the concat of the mean and LogSumExp over the residue embeddings
        if embed_type == 'fixed':
            return x

  
        
    
    def forward(self, x1, x2, x1_mask, x2_mask, task, return1 = False, return2 = False, return3 = False, 
                task_specific = False, return_cossim = False):
        
        ### Project both inputs x1 and x2 to a lower dimension through several linear + LayerNorm layers
        ### x1 and x2 are CDR-specific embeddings of antibody sequences
        if x1.shape[-1] > 6000:
            x1, x1_pos = x1[:,:,-4-self.embed_dim:-4], x1[:,:,-4:]
            x2, x2_pos = x2[:,:,-4-self.embed_dim:-4], x2[:,:,-4:]
        else:
            x1, x1_pos = x1[:,:,:-4], x1[:,:,-4:]
            x2, x2_pos = x2[:,:,:-4], x2[:,:,-4:]

        if task == 0:
            transform = self.transform2_s
        elif task == 1:
            transform = self.transform2_f

        x1 = self.project(x1)
        x2 = self.project(x2)

        if return1:
            return x1, x2

        # ADD POSITIONAL ENCODING HERE:
        x1, x2 = torch.add(x1, self.posenc(x1)), torch.add(x2, self.posenc(x2))

        # x1 = x1 - x1_pos[:,:,0:1].repeat(1,1,x1.shape[-1]) + x1_pos[:,:,2:3].repeat(1,1,x1.shape[-1])
        # x2 = x2 - x2_pos[:,:,0:1].repeat(1,1,x2.shape[-1]) + x2_pos[:,:,2:3].repeat(1,1,x2.shape[-1])

        x1 = torch.cat([x1, x1_pos], dim=-1)
        x2 = torch.cat([x2, x2_pos], dim=-1)

        # return intermediate feature 1: lower-Dim rep with positional encodings added and concatenated
        if return2:
            return x1, x2

        x1 = self.attention(x1, x1_mask, task)
        x2 = self.attention(x2, x2_mask, task)

        # return intermediate feature 2: lower-Dim rep after Transformer Encoder Block
        if return3:
            return x1, x2

        # method 1: mean of the features over the seq_len:
        if x1_mask is None:
            x1_ = torch.mean(x1, dim=1)
            x2_ = torch.mean(x2, dim=1)
        else:
            x1_ = torch.sum(x1, dim=1)
            x2_ = torch.sum(x2, dim=1)
            x1_ = torch.div(x1_, torch.unsqueeze(x1_mask.shape[-1] - torch.sum(x1_mask, dim=-1), dim=-1))
            x2_ = torch.div(x2_, torch.unsqueeze(x2_mask.shape[-1] - torch.sum(x2_mask, dim=-1), dim=-1))

        # method 2: log-sum-exp of the output features from attention layers
        x1 = torch.log(torch.sum(torch.exp(x1), dim=1))
        x2 = torch.log(torch.sum(torch.exp(x2), dim=1))

        # original normalization location
        x1, x2 = F.normalize(x1), F.normalize(x2)
        x1_, x2_ = F.normalize(x1_), F.normalize(x2_)

        # ---------------------------------------
        # concatenate mean and logsumexp
        x1 = torch.cat((x1_, x1), dim=-1)
        x2 = torch.cat((x2_, x2), dim=-1)
        # ---------------------------------------

        # x1, x2 = F.normalize(x1), F.normalize(x2)

        # Output fixed-length embedding here. These are the concat of the mean and LogSumExp over the residue embeddings
        if task_specific:
            return x1, x2

        # Take cosine similarity of two embeddings
        pred = torch.unsqueeze(self.cossim(x1, x2), dim=-1)

        if return_cossim:
            return torch.reshape(pred, (-1,)), x1, x2

        # Transform cossim for either structure or function
        pred = transform(pred)

        return torch.reshape(pred, (-1,)), x1, x2


    # def embed_dir(self, device, input_path, variable_length, task):
        # TODO - implement this?   
    #     dev = torch.device(f'cuda:{device}' if torch.cuda.is_available() else "cpu")

    #     # Get list of files or casts single file as a list
    #     if os.path.isdir(input_path): input_iter = os.listdir(input_path)
    #     elif os.path.isfile(input_path): input_iter = list(input_path)
        
    #     outputs = []
    #     for input in input_iter:
    #         with open(input, 'rb') as p:
    #             input_embed = pickle.load(p).to(dev)
    #         input_embed = torch.unsqueeze(input_embed, 0)
    #         try:
    #             assert len(input_embed.shape) == 3
    #         except:
    #             raise ValueError("input embedding should be of shape n'(CDR length) x d")

    #         # generate the abmap embedding
    #         with torch.no_grad():
    #             if variable_length:
    #                 out_feature, _ = pretrained.embed(input_embed, task=task, embed_type='variable')
    #             else:
    #                 out_feature, _ = pretrained.embed(input_embed, task=task, embed_type='fixed')
    #         out_feature = torch.squeeze(out_feature, 0)
    #         outputs.append(out_feature)
            
    #         return outputs
            


class AbMAPLSTM(nn.Module):
    def __init__(self, embed_dim = 6165, chain_type = 'H', mid_dim1 = 2048, mid_dim2 = 512, mid_dim3 = 128,
                  proj_dim = 50, hidden_dim = 50, lstm_layers = 1):
        super(AbMAPLSTM, self).__init__()
        self.lstm_layers = lstm_layers
        self.activation = nn.LeakyReLU()
        
        self.project = nn.Sequential(
                            # nn.Linear(embed_dim, mid_dim1),
                            # nn.LayerNorm(mid_dim1),
                            # self.activation,
                            nn.Linear(embed_dim, mid_dim2),
                            nn.LayerNorm(mid_dim2),
                            self.activation,
                            nn.Linear(mid_dim2, mid_dim3),
                            nn.LayerNorm(mid_dim3),
                            self.activation,
                            nn.Linear(mid_dim3, proj_dim),
                            nn.LayerNorm(proj_dim),
                            self.activation,
                            nn.Dropout(p=0.3)
                          )

        self.rnn_struc = nn.LSTM(proj_dim + 4, hidden_dim, num_layers = lstm_layers, bidirectional=True, batch_first=True)
        self.rnn_func = nn.LSTM(proj_dim + 4, hidden_dim, num_layers = lstm_layers, bidirectional=True, batch_first=True)

        self.cossim = nn.CosineSimilarity(dim=-1)

        # last layer initialization
        self.transform2_s = nn.Linear(1, 1)
        self.transform2_s.weight = torch.nn.Parameter(torch.ones(1, 1))
        self.transform2_s.bias = torch.nn.Parameter(torch.zeros(1))

        self.transform2_f = nn.Linear(1, 1)
        self.transform2_f.weight = torch.nn.Parameter(torch.ones(1, 1))
        self.transform2_f.bias = torch.nn.Parameter(torch.zeros(1))


    def recurrent(self, x, task):
        if task == 0:
            rnn = self.rnn_struc
        elif task == 1:
            rnn = self.rnn_func
        output, (h_out, c_out) = rnn(x)

        # concatenate the two batch_size x hidden_dim vectors --> batch_size x (2*hidden_dim)
        h_out = h_out.view(self.lstm_layers, 2, h_out.shape[-2], h_out.shape[-1])
        out_hidden = torch.cat([h_out[-1][0], h_out[-1][1]], dim=1)

        return out_hidden


    def forward(self, x1, x2, x1_mask, x2_mask, task):
        x1, x1_pos = x1[:,:,-2204:-4], x1[:,:,-4:]
        x2, x2_pos = x2[:,:,-2204:-4], x2[:,:,-4:]

        if task == 0:
            transform = self.transform2_s
        elif task == 1:
            transform = self.transform2_f

        x1 = self.project(x1)
        x2 = self.project(x2)

        x1 = torch.cat([x1, x1_pos], dim=-1)
        x2 = torch.cat([x2, x2_pos], dim=-1)

        x1 = self.recurrent(x1, task)
        x2 = self.recurrent(x2, task)


        x1, x2 = F.normalize(x1), F.normalize(x2)

        pred = torch.unsqueeze(self.cossim(x1, x2), dim=-1)

        pred = transform(pred)

        return torch.reshape(pred, (-1,)), x1, x2


class MultiTaskLossWrapper(nn.Module):
    def __init__(self, task_num, init_alpha, update_alpha):
        super(MultiTaskLossWrapper, self).__init__()
        self.task_num = task_num
        self.weights = nn.Parameter(torch.tensor(init_alpha), requires_grad=update_alpha) # if -2.0, alpha = e^2 = 7.xx
        self.loss_fn = nn.MSELoss()
        self.loss_s, self.loss_f = None, None
        self.alpha = None

    def forward(self, pred_s, label_s, pred_f, label_f):
        loss_s = self.loss_fn(pred_s, label_s)
        loss_f = self.loss_fn(pred_f, label_f)

        self.loss_s, self.loss_f = loss_s, loss_f

        precision_s = torch.exp(-self.weights)
        loss_s = precision_s*loss_s
        self.alpha = precision_s

        precision_f = 1/precision_s
        loss_f = precision_f*loss_f
        
        return loss_s + loss_f


class BrineyPredictor(nn.Module):
    def __init__(self, input_dim = 512, mid_dim1 = 128, mid_dim2 = 32, mid_dim3 = 8):
        super(BrineyPredictor, self).__init__()

        self.activation = nn.LeakyReLU()

        self.predictor = nn.Sequential(
                               nn.Linear(input_dim, mid_dim1),
                               self.activation,
                               nn.Linear(mid_dim1, mid_dim2),
                               self.activation,
                               nn.Linear(mid_dim2, mid_dim3),
                               self.activation,
                               # nn.Dropout(p=0.3),
                               nn.Linear(mid_dim3, 1))

    def forward(self, x):
        x = self.predictor(x)
        # x = F.relu(x)
        x = torch.exp(x)
        return torch.squeeze(x)


class BrineyPredictorAttn(nn.Module):
    def __init__(self, input_dim = 512, mid_dim1 = 128, mid_dim2 = 32, mid_dim3 = 8):
        super(BrineyPredictorAttn, self).__init__()

        self.activation = nn.LeakyReLU()

        self.predictor = nn.Sequential(
                               nn.Linear(input_dim, mid_dim1),
                               self.activation,
                               nn.Linear(mid_dim1, mid_dim2),
                               self.activation,
                               nn.Linear(mid_dim2, mid_dim3),
                               self.activation,
                               # nn.Dropout(p=0.3),
                               nn.Linear(mid_dim3, 1))

        self.enc_layer = nn.TransformerEncoderLayer(d_model=input_dim, nhead=8, batch_first=True,
                                                      dim_feedforward=input_dim)

        self.trans_enc = nn.TransformerEncoder(encoder_layer=self.enc_layer, num_layers=1)

    def attention(self, x, x_mask):
        if x_mask is None:
            output = self.trans_enc(x)
        else:
            output = self.trans_enc(x, src_key_padding_mask=x_mask)

        return output


    def forward(self, x, x_mask):
        x = self.attention(x, x_mask)
        x = torch.squeeze(self.predictor(x), dim=-1)
        x = torch.logsumexp(x, dim=-1)
        x = torch.sigmoid(x)
        return x



def init_weights(m):
    if isinstance(m, nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight, gain=0.01)

class DesautelsPredictor(nn.Module):
    def __init__(self, input_dim = 400, mid_dim = 64, embed_dim = 50, hidden_dim = 50, 
                 output_dim = 5, dropout= 0.3, baseline = False, num_heads=16):
        super(DesautelsPredictor, self).__init__()
        self.activation = nn.LeakyReLU()
        num_enc_layers = 1

        self.predictor = nn.Sequential(
                            nn.Linear(input_dim, mid_dim),
                            self.activation,
                            nn.Dropout(p=dropout),
                            nn.Linear(mid_dim, output_dim),
                          )

        self.enc_layer_H = nn.TransformerEncoderLayer(d_model=embed_dim, nhead=num_heads, batch_first=True,
                                                      dim_feedforward=hidden_dim)
        self.enc_layer_L = nn.TransformerEncoderLayer(d_model=embed_dim, nhead=num_heads, batch_first=True,
                                                      dim_feedforward=hidden_dim)

        self.attention_H = nn.TransformerEncoder(encoder_layer=self.enc_layer_H, num_layers=num_enc_layers)
        self.attention_L = nn.TransformerEncoder(encoder_layer=self.enc_layer_L, num_layers=num_enc_layers)

        # Xavier Uniform
        if baseline:
            self.attention_H.apply(init_weights)
            self.attention_L.apply(init_weights)


    def attention(self, x, chain_type):
        if chain_type == 'H':
            mha = self.attention_H
        else:
            mha = self.attention_L

        if x.shape[-1] > 6000:
            x = x[:,:,-2200:]

        output = mha(x)

        pooled_output = torch.mean(output, dim=1)

        return pooled_output

    def forward(self, h, wt_h, l, wt_l):

        if torch.isnan(h).any() or torch.isnan(wt_h).any() or torch.isnan(l).any() or torch.isnan(wt_l).any():
            raise ValueError

        h, wt_h = self.attention(h, 'H'), self.attention(wt_h, 'H')
        l, wt_l = self.attention(l, 'L'), self.attention(wt_l, 'L')

        x = torch.cat([h, h - wt_h, l, l - wt_l], dim=-1)

        output = self.predictor(x)
        return output



import os
import pickle
from tqdm import tqdm
import argparse
import random
import pickle
import numpy as np
from matplotlib import pyplot as plt
import pandas as pd
import h5py
import time

import torch
from torch.utils.data import Dataset, DataLoader
from torch.nn.utils.rnn import pad_sequence
from torch import nn



class MultiDataset(Dataset):
    def __init__(self, dset1, dset2, labels, device):
        self.dset1 = pad_sequence(dset1, batch_first=True)#.to('cuda:{}'.format(device))
        self.dset2 = pad_sequence(dset2, batch_first=True)#.to('cuda:{}'.format(device))
        self.labels = torch.FloatTensor(labels)#.to('cuda:{}'.format(device))

    def __len__(self):
        return self.labels.shape[0]

    def __getitem__(self, idx):
        return self.dset1[idx,:,:], self.dset2[idx,:,:], self.labels[idx]

def my_collate(data):
    temp1, temp2, temp3 = [], [], []
    mask1, mask2 = [], []

    for x1_path, x2_path, tms in data:
        with open(x1_path, 'rb') as f1:
            # print(x1_path)
            x1_data = pickle.load(f1)
        with open(x2_path, 'rb') as f2:
            # print(x2_path)
            x2_data = pickle.load(f2)

        x1_data.cpu()
        x2_data.cpu()

        # pre-padding step 1:
        temp1.append(torch.flip(x1_data, [0]))
        temp2.append(torch.flip(x2_data, [0]))

        mask1.append(torch.zeros(x1_data.shape[0]))
        mask2.append(torch.zeros(x2_data.shape[0]))

        # post-padding(original):
        # temp1.append(x1_data)
        # temp2.append(x2_data)

        temp3.append(tms)

        del x1_data, x2_data

    out1 = pad_sequence(temp1, batch_first=True)
    out2 = pad_sequence(temp2, batch_first=True)

    out_mask1 = pad_sequence(mask1, padding_value=1, batch_first=True)
    out_mask2 = pad_sequence(mask2, padding_value=1, batch_first=True)

    # pre-padding step 2:
    out1, out2 = torch.flip(out1, [1]), torch.flip(out2, [1])
    out_mask1, out_mask2 = torch.flip(out_mask1, [1]).type(torch.bool), torch.flip(out_mask2, [1]).type(torch.bool)

    # print(out1.shape, out_mask1.shape)
    # print(out2.shape, out_mask2.shape)
    # assert False

    ########### adding an extra dimension for features
    # out1 = torch.cat((out1, torch.zeros(1, out1.shape[1], out1.shape[2])), dim=0)
    # out2 = torch.cat((out2, torch.zeros(1, out2.shape[1], out2.shape[2])), dim=0)
    # out_mask1 = torch.cat((out_mask1, torch.zeros(out_mask1.shape[0], 1)), dim=-1)
    # out_mask2 = torch.cat((out_mask2, torch.zeros(out_mask2.shape[0], 1)), dim=-1)
    ###########


    out3 = torch.Tensor(temp3)

    return out1, out2, out3, out_mask1, out_mask2


def create_dataloader_sabdab(data_dir, batch_size, emb_type, mut_type, chain_type, region):
    # parent_dir = '../data/processed/sabdab'
    parent_dir = '/data/cb/rsingh/work/antibody/ci_data/processed/sabdab'

    embed_dir = os.path.join(parent_dir, data_dir, emb_type)

    if region == 'whole':
        data_path = os.path.join(parent_dir, 'pair_scores', 'set1_{}_pair_scores_100k.csv'.format(chain_type))
    else:
        data_path = os.path.join(parent_dir, 'pair_scores', 'set1_CDR{}_{}_pair_scores_100k.csv'.format(region, chain_type))

    data = pd.read_csv(data_path)
    data = data.iloc[:100000]

    print(data.shape[0])

    data_list = []
    for idx, row in data.iterrows():
        p1_id, p2_id, tms, rmsd = row['pdb_id1'], row['pdb_id2'], row['tm_score'], row['rmsd']

        p1 = '{}_{}_{}_k100.p'.format(p1_id, mut_type, chain_type)
        p2 = '{}_{}_{}_k100.p'.format(p2_id, mut_type, chain_type)

        # original embeddings:
        # p1 = 'sabdab_{}_{}_orig.p'.format(p1_id, chain_type)
        # p2 = 'sabdab_{}_{}_orig.p'.format(p2_id, chain_type)

        emb1 = os.path.join(embed_dir, p1)
        emb2 = os.path.join(embed_dir, p2)

        data_list.append((emb1, emb2, tms))

    # randomly split dataset into train, val, test (70-15-15):
    train_idx = int(len(data_list)*0.7)
    val_idx = int(len(data_list)*0.85)

    train_loader = DataLoader(data_list[:train_idx], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)
    val_loader = DataLoader(data_list[train_idx:val_idx], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)
    test_loader = DataLoader(data_list[val_idx:], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)

    return train_loader, val_loader, test_loader


def create_dataloader_libra(data_dir, batch_size, emb_type, mut_type, chain_type):
    # parent_dir = '../data/processed/libraseq'
    parent_dir = '/data/cb/rsingh/work/antibody/ci_data/processed/libraseq'

    embed_dir = os.path.join(parent_dir, data_dir, emb_type)
    data_path = os.path.join(parent_dir, 'libraseq_pairs_{}_Set1_100k.csv'.format(chain_type))

    df = pd.read_csv(data_path, keep_default_na=False)
    df = df.iloc[:100000]

    data_list = []
    for index, row in df.iterrows():
        p1 = '{}_{}_{}_k100.p'.format(row['source_seq1'], mut_type, chain_type)
        p2 = '{}_{}_{}_k100.p'.format(row['source_seq2'], mut_type, chain_type)

        # original embeddings:
        # p1 = '{}_{}_orig.p'.format(row['source_seq1'], chain_type)
        # p2 = '{}_{}_orig.p'.format(row['source_seq2'], chain_type)

        tms = float(row['score'])

        emb1 = os.path.join(embed_dir, p1)
        emb2 = os.path.join(embed_dir, p2)

        data_list.append((emb1, emb2, tms))

    train_idx = int(len(data_list)*0.7)
    val_idx = int(len(data_list)*0.85)

    train_loader = DataLoader(data_list[:train_idx], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)
    val_loader = DataLoader(data_list[train_idx:val_idx], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)
    test_loader = DataLoader(data_list[val_idx:], batch_size=batch_size, pin_memory=False,
                              num_workers = 4, collate_fn=my_collate, shuffle=True)

    return train_loader, val_loader, test_loader



class DesautelsDataset(Dataset):
    def __init__(self, features_H, features_L, labels):
        self.features_H = features_H
        self.features_L = features_L
        self.labels = labels   # DataFrame
        self.scores_cols = ['FoldX_Average_Whole_Model_DDG', 'FoldX_Average_Interface_Only_DDG',
                            'Statium', 'Sum_of_Rosetta_Flex_single_point_mutations',
                            'Sum_of_Rosetta_Total_Energy_single_point_mutations']
        self.label_vals = self.labels[self.scores_cols]

        # standardize label scores (x - mean) / std
        self.label_vals = (self.label_vals - self.label_vals.mean()) / self.label_vals.std()


    def __len__(self):
        return self.labels.shape[0]

    def __getitem__(self, idx):
        row = self.labels.iloc[idx]
        row_vals = self.label_vals.iloc[idx]
        feature_H = torch.FloatTensor(self.features_H[idx, :, :])
        feature_L = torch.FloatTensor(self.features_L[idx, :, :])
        feature_id = row['Antibody_ID'] # unncessary for now
        label = torch.FloatTensor(row_vals)

        return feature_H, feature_L, label

def create_dataloader_desautels(batch_size, traintest_split = 0.2, onehot=False, embed_type='bb'):
    np.random.seed(1) # for reproducability
    dataset_path = '../data/processed/desautels'

    if onehot:
        dataset_path = os.path.join(dataset_path, 'desautels_onehot.h5')
    else:
        if embed_type in ['ours_bb', 'ours_protbert', 'ours_esm1b']:
            dataset_path = os.path.join(dataset_path, 'desautels_cdrembed_{}.h5'.format(embed_type))
        else:
            dataset_path = os.path.join(dataset_path, 'desautels_{}.h5'.format(embed_type))

    score_cols = ['FoldX_Average_Whole_Model_DDG', 'FoldX_Average_Interface_Only_DDG',
                    'Statium', 'Sum_of_Rosetta_Flex_single_point_mutations',
                    'Sum_of_Rosetta_Total_Energy_single_point_mutations']

    start = time.time()
    with h5py.File(dataset_path, "r") as f:
        dataset_H_all, dataset_L_all = f['H'], f['L']

        # load the label:
        label_path = os.path.join(dataset_path, "Desautels_m396_seqs.csv")
        invalid_idx_path = os.path.join(dataset_path, "invalid_idx.p")
        labels_df = pd.read_csv(label_path)

        # identify all rows with NaN score values, and remove them
        extra_invalid = labels_df[labels_df['Statium'].isnull()].index.tolist()

        with open(invalid_idx_path, 'rb') as f:
            invalid_idx2 = pickle.load(f)

        invalid_idx = sorted(list(set(extra_invalid) | set(invalid_idx2)))

        all_idxs = list(range(len(dataset_H_all)))
        valid_idxs = [i for i in all_idxs if i not in invalid_idx]

        # FOR TESTING ONLY!
        # valid_idxs = valid_idxs[:1000]
        
        if embed_type == 'bb':
            dataset_H_all = dataset_H_all[:, :, -2200:]
            dataset_L_all = dataset_L_all[:, :, -2200:]

        dataset_H_all = dataset_H_all[valid_idxs, :, :]
        dataset_L_all = dataset_L_all[valid_idxs, :, :]
        labels_df = labels_df.iloc[valid_idxs]
        labels_df = labels_df.reset_index(drop=True)

        # check for NANs:
        if np.any(np.isnan(dataset_H_all)):
            print("Dataset H contains NaNs!")
            raise ValueError
        if np.any(np.isnan(dataset_L_all)):
            print("Dataset L contains NaNs!")
            raise ValueError

        # valid_idx = valid_idx[:1000] # For testing. COMMENT OUT when running actual experiment!
        # dataset_H_all, dataset_L_all, labels_df = dataset_H_all[:2000], dataset_L_all[:2000], labels_df.head(2000)
        print(f"Dataset Size: {len(dataset_H_all)}. (1000 is for testing)")

        assert dataset_H_all.shape[0] == dataset_L_all.shape[0] == labels_df.shape[0]

        # calculate the splits here:
        dataset_len = dataset_H_all.shape[0]
        train_size = int(traintest_split * dataset_len)
        all_idx = list(range(dataset_len))
        np.random.shuffle(all_idx)
        trainval_idx = all_idx[:train_size]
        trainval_split = int(0.9*train_size)

        train_idx, val_idx = sorted(trainval_idx[:trainval_split]), sorted(trainval_idx[trainval_split:])
        test_idx = sorted(all_idx[train_size:])
        
        # split the datasets into train / val / test
        features_H_train, features_L_train = dataset_H_all[train_idx, :, :], dataset_L_all[train_idx, :, :]
        labels_df_train = labels_df.iloc[train_idx].reset_index(drop=True)
        features_H_val, features_L_val = dataset_H_all[val_idx, :, :], dataset_L_all[val_idx, :, :]
        labels_df_val = labels_df.iloc[val_idx].reset_index(drop=True)
        features_H_test, features_L_test = dataset_H_all[test_idx, :, :], dataset_L_all[test_idx, :, :]
        labels_df_test = labels_df.iloc[test_idx].reset_index(drop=True)

        end = time.time()
        print("Time it took to load the H5 files: {:.2f} seconds".format(end-start))

        train_data = DesautelsDataset(features_H_train, features_L_train, labels_df_train)
        val_data = DesautelsDataset(features_H_val, features_L_val, labels_df_val)
        test_data = DesautelsDataset(features_H_test, features_L_test, labels_df_test)

        print("Train/Test split: {}, {}".format(len(train_data)+len(val_data), len(test_data)))



    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True, num_workers = 4)
    val_loader = DataLoader(val_data, batch_size=batch_size, shuffle=True, num_workers = 4)
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True, num_workers = 4)

    return train_loader, val_loader, test_loader


import torch
import abmap.base_config as base_config
from dscript.alphabets import Uniprot21
import re



def reload_models_to_device(device_num=0, plm_type='beplerberger'):
    base_config.device = "cuda:{}".format(device_num)
    device = torch.device(base_config.device if torch.cuda.is_available() else "cpu")
    print(f'{plm_type} loaded to {device}')

    ######### Bepler & Berger ########
    if plm_type == 'beplerberger':
        from dscript import pretrained as ds_pre
        global bb_model
        bb_model = ds_pre.get_pretrained("lm_v1")
        bb_model = bb_model.to(device)
        bb_model.eval()

    ######### ESM-1b or ESM-2 ##########
    elif plm_type in ('esm1b', 'esm2'):
        global esm_model, esm_batch_converter
        if plm_type == 'esm1b':
            esm_model, esm_alphabet = torch.hub.load("facebookresearch/esm:main", "esm1b_t33_650M_UR50S")
        else:
            esm_model, esm_alphabet = torch.hub.load("facebookresearch/esm:main", "esm2_t33_650M_UR50D")
        esm_batch_converter = esm_alphabet.get_batch_converter()
        esm_model = esm_model.to(device)
        esm_model.eval()
    
    ######### TAPE ##########
    elif plm_type == 'tape':
        global tape_model, tape_tokenizer
        from tape import ProteinBertModel, TAPETokenizer
        tape_model = ProteinBertModel.from_pretrained('bert-base')
        tape_tokenizer = TAPETokenizer(vocab='iupac')
        tape_model = tape_model.to(device)
        tape_model.eval()

    ######### ProtBert #########
    elif plm_type == 'protbert':
        from transformers import BertModel, BertTokenizer
        global protbert_tokenizer, protbert_model
        protbert_tokenizer = BertTokenizer.from_pretrained("Rostlab/prot_bert", do_lower_case=False )
        protbert_model = BertModel.from_pretrained("Rostlab/prot_bert")
        protbert_model = protbert_model.to(device)
        protbert_model.eval()

    else:
        raise ModuleNotFoundError(f"PLM {plm_type} not found...")


def embed_sequence(sequence, embed_type = "beplerberger", embed_device = None, embed_model=None):
    if embed_device is None:
        device = torch.device(base_config.device if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(embed_device)

    if embed_type == "beplerberger":
        from dscript.alphabets import Uniprot21
        # print("using Bepler & Berger's Embedding...")

        with torch.no_grad():
            alphabet = Uniprot21()
            es = torch.from_numpy(alphabet.encode(sequence.encode('utf-8')))
            x = es.long().unsqueeze(0)
            x = x.to(device)

            if embed_model is None:
                z = bb_model.transform(x)
            else:
                z = embed_model.transform(x)

            return z.detach().cpu()[0]


    elif embed_type == "esm1b" or embed_type == "esm2":
        # print("using FAIR's esm-1b...")

        data = [
            ("seq", sequence),
        ]
        batch_labels, batch_strs, batch_tokens = esm_batch_converter(data)
        # think it adds beginning and end tokens in the conversion process: ex) seq length 16 --> 18

        batch_tokens = batch_tokens.to(device)
        # Extract per-residue representations (on CPU)
        with torch.no_grad():
            if embed_model is None:
                results = esm_model(batch_tokens, repr_layers=[33], return_contacts=True)
            else:
                results = embed_model(batch_tokens, repr_layers=[33], return_contacts=True)

        out_temp = results["representations"][33]
        embedding = out_temp[0,1:-1,:] # remove start and end tokens

        return embedding #(out_temp[0], out_temp[1])


    elif embed_type == "tape":
        assert embed_device is None and embed_model is None # not implemented yet
        # print("using Berkeley's TAPE...")
        model = tape_model.eval().to(device)

        token_ids = torch.tensor([tape_tokenizer.encode(sequence)]) #.to(device)
        # token_ids = token_ids.to()
        with torch.no_grad():
            if embed_model is None:
                output = model(token_ids.to(device))
            else:
                output = embed_model(token_ids.to(device))
        embedding = output[0][0,1:-1,:] # remove start and end tokens

        return embedding


    elif embed_type == "dscript":
        from dscript.alphabets import Uniprot21
        assert embed_device is None and embed_model is None # not implemented yet
        # print("using D-Script Embedding (Bepler&Berger + projection)...")
        with torch.no_grad():
            alphabet = Uniprot21()
            es = torch.from_numpy(alphabet.encode(sequence.encode('utf-8')))
            x = es.long().unsqueeze(0)
            x = x.to(device)
            z = bb_model.transform(x)
            # return z.cpu()[0]

        model = ds_model.eval().to(device)
        embedding = model.embedding(z)[0]

        return embedding #.to()

    elif embed_type == 'protbert':

        sequence_spaced = " ".join(sequence)
    
        with torch.no_grad():
            encoded_input = protbert_tokenizer(sequence_spaced, return_tensors='pt')
            # encoded_input = encoded_input.to(device)
            # print(encoded_input)
            for key in encoded_input.keys():
                encoded_input[key] = encoded_input[key].to(device)
            output = protbert_model(**encoded_input)
            feature = output.last_hidden_state[0][1:-1,:]

        return feature




import torch
import os, sys
import pickle
from tqdm import tqdm
import time
import argparse
import h5py
import numpy as np

from torch import nn
from torch import optim as opt
from torch import autograd
from torch.optim.lr_scheduler import MultiStepLR
import torch.nn.functional as F

from sklearn.linear_model import Ridge
from sklearn.model_selection import KFold

sys.path.append('..')
from model import AbMAPAttn, DesautelsPredictor, MultiTaskLossWrapper
from dataloader import create_dataloader_desautels
from utils import evaluate_spearman
from embed import reload_models_to_device
from protein_embedding import ProteinEmbedding


pred_cols = ['FoldX_Average_Whole_Model_DDG', 'FoldX_Average_Interface_Only_DDG',
            'Statium', 'Sum_of_Rosetta_Flex_single_point_mutations',
            'Sum_of_Rosetta_Total_Energy_single_point_mutations']

all_aas = ['A', 'G', 'V', 'I', 'L', 'F', 'P', 'Y', 'M', 'T',
           'S', 'H', 'N', 'Q', 'W', 'R', 'K', 'D', 'E', 'C', 'X']

H_cdr_idx = [23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107]
L_cdr_idx = [22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 46, 47, 48, 49, 50, 51, 52, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97]

def train_model_onehot(args, device_num, num_epochs, batch_size, exec_type, lr, traintest_split):
    """
    This function trains the ddG prediction network using 
    the one-hot representations of input amino acids
    """

    device = torch.device("cuda:{}".format(device_num) if torch.cuda.is_available() else "cpu")

    # load the model/optimizer
    model = DesautelsPredictor(input_dim=84, mid_dim = 16, embed_dim = 21, hidden_dim=21, num_heads=7).to(device)
    loss_fn = nn.MSELoss()
    optimizer = opt.SGD(model.parameters(), lr=lr)
    scheduler = MultiStepLR(optimizer, milestones=[50, 80], gamma=0.1)

    print("Building dataloader...")
    # load the dataloader
    train_loader_des, val_loader_des, test_loader_des = create_dataloader_desautels(batch_size=batch_size, 
                                                            traintest_split = traintest_split, onehot=True)
    print("Done building dataloader!")

    # load WT features:
    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/WT_feats/WT_H_onehot.p', 'rb') as f:
        WT_H_feat = pickle.load(f).to(device)
    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/WT_feats/WT_L_onehot.p', 'rb') as f:
        WT_L_feat = pickle.load(f).to(device)


    X_train, y_train = [],[]
    X_test, y_test = [],[]

    if exec_type == 'train':
        # with autograd.detect_anomaly():
        for i in range(num_epochs):

            # train the model
            model.train()
            total_loss = 0

            for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(train_loader_des):

                # predict the label using the model
                feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

                # # create the WT features batch:
                WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
                WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

                # change to floats:
                feat_H_batch = feat_H_batch.float()
                feat_L_batch = feat_L_batch.float()
                WT_H_feat_batch = WT_H_feat_batch.float()
                WT_L_feat_batch = WT_L_feat_batch.float()

                pred_batch = model(feat_H_batch, WT_H_feat_batch, feat_L_batch, WT_L_feat_batch)

                # update the parameters (loss, backprop, etc.)
                loss = loss_fn(pred_batch, label_batch)
                total_loss += loss.item()
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

                if k == len(train_loader_des) - 1:
                    print("TRAIN Epoch {}, Batch {}, Loss: {}".format(i, k, total_loss/(k+1)))

            scheduler.step()

            # print("********************")

            # validate the model
            model.eval()
            total_loss = 0

            for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(val_loader_des):

                # predict the label using the model
                feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

                # # create the WT features batch:
                WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
                WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

                # change to floats:
                feat_H_batch = feat_H_batch.float()
                feat_L_batch = feat_L_batch.float()
                WT_H_feat_batch = WT_H_feat_batch.float()
                WT_L_feat_batch = WT_L_feat_batch.float()

                with torch.no_grad():
                    pred_batch = model(feat_H_batch, WT_H_feat_batch, feat_L_batch, WT_L_feat_batch)


                # update the parameters (loss, backprop, etc.)
                loss = loss_fn(pred_batch, label_batch)
                total_loss += loss.item()


                if k == len(val_loader_des) - 1:
                    print("VAL Epoch {}, Batch {}, Loss: {}".format(i, k, total_loss/(k+1)))

            print("----------------------------------------------")

    # feed the model through the test set
    model.eval()
    losses = []
    labels_total, preds_total = [], []

    for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(test_loader_des): #, total=len(test_loader_des):

        # predict the label using the model
        feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

        # # create the WT features batch:
        WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
        WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

        # change to floats:
        feat_H_batch = feat_H_batch.float()
        feat_L_batch = feat_L_batch.float()
        WT_H_feat_batch = WT_H_feat_batch.float()
        WT_L_feat_batch = WT_L_feat_batch.float()

        with torch.no_grad():
            pred_batch = model(feat_H_batch, WT_H_feat_batch, feat_L_batch, WT_L_feat_batch)

        #-----------------------------------------------------

        labels_total.append(label_batch)
        preds_total.append(pred_batch)

        # update the parameters (loss, backprop, etc.)
        loss = loss_fn(pred_batch, label_batch)
        losses.append(loss.item())

    save_dict = dict()

    # Calculate Average Loss over the TEST SET:
    print("TEST SET Average LOSS: {}".format(sum(losses)/len(losses)))

    # calculate spearman rank for each value:
    labels_total = torch.cat(labels_total, dim=0)
    preds_total = torch.cat(preds_total, dim=0)

    for j in range(labels_total.shape[-1]):
        spear = evaluate_spearman(preds_total[:, j], labels_total[:, j])
        print("Split: {}, Spearman Rank Scores for {}: {}".format(traintest_split, pred_cols[j], spear))
        save_dict[pred_cols[j]] = (preds_total[:, j], labels_total[:, j])

    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/plots/{}_spl{}.p'.format(args.logname, traintest_split), 'wb') as p:
        pickle.dump(save_dict, p)



def train_model(args, device_num, num_epochs, batch_size, exec_type, lr, ridge, traintest_split, embed_type):
    """
    This function executes either:
    a) Ridge Regression using our model's fixed-length embeddings

    or

    b) Training the ddG prediction model using either:
        1) our model's representations
        2) ESM-1b representations
        3) ProtBert representations

    """

    device = torch.device("cuda:{}".format(device_num) if torch.cuda.is_available() else "cpu")

    embed_size_dict = {'bb': 2200, 'esm1b': 1280, 'protbert': 1024}
    if embed_type in ['bb', 'esm1b', 'protbert']:
        embed_size = embed_size_dict[embed_type]
        mid_dim = 128*2 if embed_type == 'bb' else 64*2

    if embed_type == 'ours_bb':
        embed_size = 256
        mid_dim = 32*2

        # load the pre-trained multi-task training model:
        pretrained_H = AbMAPAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        pretrained_L = AbMAPAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        root_path = '/net/scratch3.mit.edu/scratch3-3/chihoim/model_ckpts'
        H_best_path = os.path.join(root_path, '091522_bb_newnum_H', 'beplerberger_epoch50.pt')
        L_best_path = os.path.join(root_path, '091522_bb_newnum_L', 'beplerberger_epoch50.pt')

    if embed_type == 'ours_protbert':
        embed_size = 256
        mid_dim = 32*2

        # load the pre-trained multi-task training model:
        pretrained_H = AbMAPAttn(embed_dim=1024, mid_dim2=512, mid_dim3=256,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        pretrained_L = AbMAPAttn(embed_dim=1024, mid_dim2=512, mid_dim3=256,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        root_path = '/net/scratch3.mit.edu/scratch3-3/chihoim/model_ckpts'
        H_best_path = os.path.join(root_path, '101722_abnet_protbert_H_whole', 'protbert_epoch50.pt')
        L_best_path = os.path.join(root_path, '101722_abnet_protbert_L_whole', 'protbert_epoch50.pt')

    if embed_type == 'ours_esm1b':
        embed_size = 256
        mid_dim = 32*2

        # load the pre-trained multi-task training model:
        pretrained_H = AbMAPAttn(embed_dim=1280, mid_dim2=512, mid_dim3=256,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        pretrained_L = AbMAPAttn(embed_dim=1280, mid_dim2=512, mid_dim3=256,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
        root_path = '/net/scratch3.mit.edu/scratch3-3/chihoim/model_ckpts'
        H_best_path = os.path.join(root_path, '103122_abnet_esm1b_Hwhole', 'esm1b_epoch50.pt')
        L_best_path = os.path.join(root_path, '103122_abnet_esm1b_Lwhole', 'esm1b_epoch50.pt')

    if embed_type in ['ours_bb', 'ours_protbert', 'ours_esm1b']:
        checkpoint_H = torch.load(H_best_path, map_location=device)
        checkpoint_L = torch.load(L_best_path, map_location=device)
        pretrained_H.load_state_dict(checkpoint_H['model_state_dict'])
        pretrained_L.load_state_dict(checkpoint_L['model_state_dict'])
        pretrained_H.eval()
        pretrained_L.eval()


    # load the predictor model/optimizer
    model = DesautelsPredictor(input_dim=embed_size*4, mid_dim = mid_dim, embed_dim = embed_size, 
                               hidden_dim=embed_size, num_heads=16).to(device)
    loss_fn = nn.MSELoss()
    optimizer = opt.SGD(model.parameters(), lr=lr)
    scheduler = MultiStepLR(optimizer, milestones=[50, 80], gamma=0.1)

    print("Building dataloader...")
    # load the dataloader
    train_loader_des, val_loader_des, test_loader_des = create_dataloader_desautels(batch_size=batch_size, 
                                             traintest_split = traintest_split, embed_type=args.embed_type)
    print("Done building dataloader!")

    # load WT features:
    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/WT_feats/WT_H_{}.p'.format(args.embed_type), 'rb') as f:
        WT_H_feat = pickle.load(f).to(device)
    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/WT_feats/WT_L_{}.p'.format(args.embed_type), 'rb') as f:
        WT_L_feat = pickle.load(f).to(device)


    X_train, y_train = [],[]
    X_test, y_test = [],[]

    if exec_type == 'train':
        # with autograd.detect_anomaly():
        for i in range(num_epochs):

            # train the model
            model.train()
            total_loss = 0

            for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(train_loader_des):

                # predict the label using the model
                feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

                # # create the WT features batch:
                WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
                WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

                # for RIDGE REGRESSION:
                with torch.no_grad():
                    if ridge:
                        task = 0
                        feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=task, task_specific=True)
                        feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=task, task_specific=True)
                        feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=task, task_specific=True)
                        feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=task, task_specific=True)
                        final_input_feat = torch.cat([feat_h, feat_wt_h, feat_l, feat_wt_l], dim=-1)

                        if torch.isnan(final_input_feat).any():
                            print("Train Tensor Contains Nan!")

                        if torch.isnan(label_batch).any():
                            print("Train Label Contains Nan!")

                        X_train.append(final_input_feat.detach().cpu())
                        y_train.append(label_batch.detach().cpu())
                        continue

                    # print(feat_h.shape, feat_wt_h.shape, feat_l.shape, feat_wt_l.shape)

                    if embed_type in ['ours_bb', 'ours_protbert', 'ours_esm1b']:
                        # project all feature dimensions down to 256 (for our model):
                        feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=0, return2=True)
                        feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=0, return2=True)
                        feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=0, return2=True)
                        feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=0, return2=True)

                    if embed_type in ['bb', 'esm1b', 'protbert']:
                        # for other embeddings (ESM, ProtBERT, etc.)
                        feat_h = feat_H_batch
                        feat_wt_h = WT_H_feat_batch
                        feat_l = feat_L_batch
                        feat_wt_l = WT_L_feat_batch


                pred_batch = model(feat_h, feat_wt_h, feat_l, feat_wt_l)

                # update the parameters (loss, backprop, etc.)
                loss = loss_fn(pred_batch, label_batch)
                total_loss += loss.item()
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

                if k == len(train_loader_des) - 1:
                    print("TRAIN Epoch {}, Batch {}, Loss: {}".format(i, k, total_loss/(k+1)))

            

            # print("********************")

            # validate the model
            model.eval()
            total_loss = 0

            for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(val_loader_des):

                # predict the label using the model
                feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

                # # create the WT features batch:
                WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
                WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

                with torch.no_grad():

                    # for RIDGE REGRESSION:
                    if ridge:
                        task = 0
                        feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=task, task_specific=True)
                        feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=task, task_specific=True)
                        feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=task, task_specific=True)
                        feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=task, task_specific=True)
                        final_input_feat = torch.cat([feat_h, feat_wt_h, feat_l, feat_wt_l], dim=-1)

                        if torch.isnan(final_input_feat).any():
                            print("VAL Tensor Contains Nan!")

                        if torch.isnan(label_batch).any():
                            print("VAL Label Contains Nan!")

                        X_train.append(final_input_feat.detach().cpu())
                        y_train.append(label_batch.detach().cpu())
                        continue

                    if embed_type in ['ours_bb', 'ours_protbert', 'ours_esm1b']:
                        # project all feature dimensions down to 256 (for our model):
                        feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=0, return2=True)
                        feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=0, return2=True)
                        feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=0, return2=True)
                        feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=0, return2=True)

                    if embed_type in ['bb', 'esm1b', 'protbert']:
                        # for other embeddings (ESM, ProtBERT, etc.)
                        feat_h = feat_H_batch
                        feat_wt_h = WT_H_feat_batch
                        feat_l = feat_L_batch
                        feat_wt_l = WT_L_feat_batch


                    pred_batch = model(feat_h, feat_wt_h, feat_l, feat_wt_l)

                # update the parameters (loss, backprop, etc.)
                loss = loss_fn(pred_batch, label_batch)
                total_loss += loss.item()

                if k == len(val_loader_des) - 1:
                    print("VAl Epoch {}, Batch {}, Loss: {}".format(i, k, total_loss/(k+1)))

            print("----------------------------------------------")

            if ridge:
                break
            
            scheduler.step()


    # FIT the RIDGE REGRESSION MODEL HERE:
    # X, y = torch.cat(X_train, dim=0).numpy(), torch.cat(y_train, dim=0).numpy()
    # assert len(X) == len(y)
    # kf = KFold(n_splits=5, shuffle=True)
    # alphas = [0.1, 0.5, 1, 5, 10]
    # print("Cross Validation over alpha values: {}".format(alphas))
    # for i, (train_index, val_index) in enumerate(kf.split(X)):
    #     X_train, X_val = X[train_index], X[val_index]
    #     y_train, y_val = y[train_index], y[val_index]
    #     alpha = alphas[i]
    #     clf = Ridge(alpha=alpha)
    #     clf.fit(X_train, y_train)

    #     # prediction on Val:
    #     sc = clf.score(X_val, y_val)
    #     print("Alpha: {}, Ridge Regression Score: {}".format(alpha, sc))

    # sys.exit("Done Testing Cross Validation")


    # feed the model through the test set
    model.eval()
    losses = []
    labels_total, preds_total = [], []

    for k, (feat_H_batch, feat_L_batch, label_batch) in enumerate(test_loader_des): #, total=len(test_loader_des):

        # predict the label using the model
        feat_H_batch, feat_L_batch, label_batch = feat_H_batch.to(device), feat_L_batch.to(device), label_batch.to(device)

        # # create the WT features batch:
        WT_H_feat_batch = WT_H_feat.repeat(feat_H_batch.shape[0], 1, 1)
        WT_L_feat_batch = WT_L_feat.repeat(feat_L_batch.shape[0], 1, 1)

        with torch.no_grad():

            # for RIDGE REGRESSION
            if ridge:
                task = 0
                feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=task, task_specific=True)
                feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=task, task_specific=True)
                feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=task, task_specific=True)
                feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=task, task_specific=True)
                final_input_feat = torch.cat([feat_h, feat_wt_h, feat_l, feat_wt_l], dim=-1)

                if torch.isnan(final_input_feat).any():
                    print("Test Tensor Contains Nan!")

                if torch.isnan(label_batch).any():
                    print("Test Label Contains Nan!")

                X_test.append(final_input_feat.detach().cpu())
                y_test.append(label_batch.detach().cpu())
                continue

            if embed_type in ['ours_bb', 'ours_protbert', 'ours_esm1b']:
                # project all feature dimensions down to 256 (for our model):
                feat_h, _ = pretrained_H(feat_H_batch, feat_H_batch, None, None, task=0, return2=True)
                feat_wt_h, _ = pretrained_H(WT_H_feat_batch, WT_H_feat_batch, None, None, task=0, return2=True)
                feat_l, _ = pretrained_L(feat_L_batch, feat_L_batch, None, None, task=0, return2=True)
                feat_wt_l, _ = pretrained_L(WT_L_feat_batch, WT_L_feat_batch, None, None, task=0, return2=True)

            if embed_type in ['bb', 'esm1b', 'protbert']:
                # for other embeddings (ESM, ProtBERT, etc.)
                feat_h = feat_H_batch
                feat_wt_h = WT_H_feat_batch
                feat_l = feat_L_batch
                feat_wt_l = WT_L_feat_batch

            pred_batch = model(feat_h, feat_wt_h, feat_l, feat_wt_l)

        #-----------------------------------------------------

        labels_total.append(label_batch)
        preds_total.append(pred_batch)

        # update the parameters (loss, backprop, etc.)
        loss = loss_fn(pred_batch, label_batch)
        losses.append(loss.item())

    save_dict = dict()

    # perform RIDGE REGRESSION HERE:
    if ridge:
        X_train, y_train = torch.cat(X_train, dim=0).numpy(), torch.cat(y_train, dim=0).numpy()
        X_test, y_test = torch.cat(X_test, dim=0).numpy(), torch.cat(y_test, dim=0).numpy()
        clf = Ridge(alpha=0.5)
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)
        for j in range(y_pred.shape[-1]):
            spear = evaluate_spearman(y_pred[:, j], y_test[:, j])
            print("Spearman Rank Scores for {}: {}".format(pred_cols[j], spear))
            save_dict[pred_cols[j]] = (y_pred[:, j], y_test[:, j])
        test_score = clf.score(X_test, y_test)
        print("Train/Test Split: {}, Ridge Regression Score on Test: {:.2f}".format(traintest_split, test_score))

        with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/plots/{}_spl{}.p'.format(args.logname, traintest_split), 'wb') as p:
            pickle.dump(save_dict, p)

        return


    # Calculate Average Loss over the TEST SET:
    print("TEST SET Average LOSS: {}".format(sum(losses)/len(losses)))

    # calculate spearman rank for each value:
    labels_total = torch.cat(labels_total, dim=0)
    preds_total = torch.cat(preds_total, dim=0)

    for j in range(labels_total.shape[-1]):
        spear = evaluate_spearman(preds_total[:, j], labels_total[:, j])
        print("Split: {}, Spearman Rank Scores for {}: {}".format(traintest_split, pred_cols[j], spear))
        save_dict[pred_cols[j]] = (preds_total[:, j], labels_total[:, j], spear)

    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/plots/{}_spl{}.p'.format(args.logname, traintest_split), 'wb') as p:
        pickle.dump(save_dict, p)


def calc_topk_overlap(args, traintest_split, r1, r2):
    plots_path = '/data/cb/rsingh/work/antibody/ci_data/processed/desautels/plots'
    with open(os.path.join(plots_path, '{}_spl{}.p'.format(args.logname, traintest_split)), 'rb') as p:
        results = pickle.load(p)

    all_overlaps = []
    for key in results.keys():
        preds_total, labels_total = results[key][:2]

        k1 = int(r1*len(labels_total))
        k2 = int(r2*len(preds_total))

        if not torch.is_tensor(labels_total):
            labels_total = torch.Tensor(labels_total)
            preds_total = torch.Tensor(preds_total)

        labels_topk_idx = torch.topk(labels_total, k1).indices.detach().cpu()
        preds_topk_idx = torch.topk(preds_total, k2).indices.detach().cpu()

        overlap = np.intersect1d(labels_topk_idx, preds_topk_idx)
        print("Score {}. Overlap Ratio: {}/{} = {:.2f}".format(key, len(overlap), k1, len(overlap)/k1))
        all_overlaps.append(len(overlap)/k1)

    print("Average Overlap: {:.2f}".format(np.mean(all_overlaps)))




def preprocess_raw_data():
    """
    Take the raw Desautels dataset and generate the heavy and light chain sequence fasta files
    """

    df = pd.read_csv('../desaut_things/Desautels_m396_seqs.csv')

    for chain in ['heavy_chains', 'light_chains']:
        list_save = []
        for idx, row in df.iterrows():
            list_save.append('>'+row['Antibody_ID']+'\n')

            if chain == 'heavy_chains':
                seq = row['Antibody_Sequence'][:245]
            else:
                seq = row['Antibody_Sequence'][245:]

            if idx != -1:
                seq += '\n'
            list_save.append(seq)

        with open('/data/cb/rsingh/work/antibody/ci_data/raw/desautels/{}.fasta'.format(chain), 'w') as f:
            f.writelines(list_save)

    print("Divided the chains into heavy and light chains and saved as fasta files.")


def generate_hdf5_onehot(args, file_name):
    reload_models_to_device(args.device_num)

    with h5py.File('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/{}.h5'.format(file_name), 'w') as g:

        dset_h = g.create_dataset('H', (89263, 32, 21))
        dset_l = g.create_dataset('L', (89263, 29, 21))

        for chain_type in ['H', 'L']:
            if chain_type == 'H':
                dset = dset_h
                list_path = '/data/cb/rsingh/work/antibody/ci_data/raw/desautels/heavy_chains.fasta'
                cdr_idx = H_cdr_idx
            else:
                dset = dset_l
                list_path = '/data/cb/rsingh/work/antibody/ci_data/raw/desautels/light_chains.fasta'
                cdr_idx = L_cdr_idx
            
            
            with open(list_path, 'r') as f:
                seqs_list = f.read().splitlines()

            k=50
            for i in tqdm(range(0, len(seqs_list), 2)):
                seq = seqs_list[i+1]

                seq_num = torch.Tensor([all_aas.index(char) for char in seq])
                seq_num_cdr = seq_num[cdr_idx].type(torch.int64)
                one_hot = F.one_hot(seq_num_cdr, num_classes=len(all_aas))
                dset[int(i/2),:,:] = one_hot

                    # continue


                # prot = ProteinEmbedding(seq, chain_type, dev=1, fold=1)
                # prot.embed_seq('beplerberger')
                # prot.create_cdr_mask()
                # kmut_matr_h = prot.create_kmut_matrix(num_muts=k, embed_type='beplerberger')
                # cdr_embed = prot.create_cdr_embedding(kmut_matr_h, sep = False, mask = True)


                # dset[int(i/2), :, :] = cdr_embed
                # del cdr_embed

def generate_h5_from_tens(args):
    embed_dir_H = '/data/cb/rsingh/work/antibody/ci_data/processed/desautels/desautels_embeds_{}_H'.format(args.embed_type)
    embed_dir_L = '/data/cb/rsingh/work/antibody/ci_data/processed/desautels/desautels_embeds_{}_L'.format(args.embed_type)
    file_name = 'desautels_cdrembed_{}'.format(args.embed_type)
    file_name = 'desautels_bb'

    # valid_nums = list(range(89263))
    # invalid_nums_1 = [9, 79, 979, 8979]
    embed_size_dict = {'bb': 6165, 'ours_bb': 6165, 'ours_esm1b': 1280, 'ours_protbert': 1024}

    invalid_nums = []

    H_CDR_LEN, L_CDR_LEN = 245, 213
    embed_size = embed_size_dict[args.embed_type] + 4
    embed_size = 6165
    # Check if all antibodies have the above CDR lens! (need to do this just once)
    total_size = len(os.listdir(embed_dir_H))
    # total_size = 5000 # FOR TESTING!
    for num in tqdm(range(total_size)):
        with open(os.path.join(embed_dir_H, '{}.p'.format(num)), 'rb') as p:
            tens_H = pickle.load(p)
        with open(os.path.join(embed_dir_L, '{}.p'.format(num)), 'rb') as p:
            tens_L = pickle.load(p)
        if tens_H.shape[0] != H_CDR_LEN or tens_L.shape[0] != L_CDR_LEN:
            invalid_nums.append(num)
            continue
        if tens_H is None or tens_L is None:
            invalid_nums.append(num)

    # invalid_nums = list(set(invalid_nums_1) | set(invalid_nums_2))

    filtered_size = total_size - len(invalid_nums)
    print(invalid_nums)

    with open('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/invalid_idx.p', 'wb') as f:
        pickle.dump(invalid_nums, f)

    with h5py.File('/data/cb/rsingh/work/antibody/ci_data/processed/desautels/{}.h5'.format(file_name), 'w') as g:

        count = 0
        dset_h = g.create_dataset('H', (total_size, H_CDR_LEN, embed_size))
        dset_l = g.create_dataset('L', (total_size, L_CDR_LEN, embed_size))
        for num in tqdm(range(total_size)):
            if num in invalid_nums:
                continue
            with open(os.path.join(embed_dir_H, '{}.p'.format(num)), 'rb') as p:
                tens_H = pickle.load(p)
            with open(os.path.join(embed_dir_L, '{}.p'.format(num)), 'rb') as p:
                tens_L = pickle.load(p)

            try:
                dset_h[num, :, :] = tens_H
                dset_l[num, :, :] = tens_L
                count += 1
            except:
                print(f"idx {num} invalid! Something is wrong.")

        assert count == filtered_size

        print(f"# of valid entries for {args.embed_type}: {count}")
        print("Desautels H dataset shape:", dset_h.shape)
        print("Desautels L dataset shape:", dset_l.shape)

    # print("Additionally {} are invalid".format(additional_invalid))




def init(dev_num, c_type, emb_type):
    global device_n, chain_type, embed_type
    chain_type = c_type
    device_n = dev_num
    embed_type = emb_type
    device = torch.device("cuda:{}".format(device_n) if torch.cuda.is_available() else "cpu")
    reload_models_to_device(device_n)

def work_parallel(pair):
    dev, seq = pair
    prot = ProteinEmbedding(seq, chain_type, embed_device=f"cuda:{device_n}", dev=dev, fold=device_n)

    prot.embed_seq(embed_type)
    embedding = prot.embedding
    return (dev, embedding.detach().cpu().numpy())

    try:
        prot.create_cdr_mask()
    except:
        return (None, None)

    k=50
    cdr_embed = prot.create_cdr_specific_embedding(embed_type=embed_type, k=k)

    return (dev, cdr_embed.detach().cpu().numpy())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('--mode', type=str, help='which code path to run.')
    parser.add_argument('--chain_type', type=str, help='Which chain are we processing', choices=['H', 'L'])
    parser.add_argument('--is_onehot', type=bool, help='Which model are we running: ours/onehot')
    parser.add_argument('--device_num', type=int, help='GPU device we are using.', default=0)
    parser.add_argument('--embed_type', type=str, help='Embedding Type')
    parser.add_argument('--logname', type=str, help='Name of the file for saving')

    args = parser.parse_args()

    if args.mode == '1_preprocess_raw_data':
        preprocess_raw_data()

    elif args.mode == '2_generate_features_onehot':
        generate_hdf5_onehot(args, file_name='desautels_onehot')

    elif args.mode == '3_generate_hdf5':
        # load the sequences:
        c_type = args.chain_type
        des_proc = '/data/cb/rsingh/work/antibody/ci_data/processed/desautels'
        des_raw = '/data/cb/rsingh/work/antibody/ci_data/raw/desautels'

        if c_type == 'H':
            list_path = os.path.join(des_raw, 'heavy_chains.fasta')
        else:
            list_path = os.path.join(des_raw, 'light_chains.fasta')

        with open(list_path, 'r') as f:
            seqs_list = f.read().splitlines()

        input_seqs = []
        for i in range(0, len(seqs_list), 2):
            input_seqs.append((int(i/2), seqs_list[i+1]))

        #TESTING with few input seqs:
        # input_seqs = input_seqs[:1000]

        import multiprocessing as mp
        from multiprocessing import get_context
        embeds_result = []

        with get_context('spawn').Pool(processes=16, initializer=init, initargs=[args.device_num, c_type, args.embed_type]) as p:
            for result in tqdm(p.imap_unordered(work_parallel, input_seqs, chunksize=4), total=len(input_seqs)):
                idx_, emb = result
                with open(os.path.join(des_proc, f'desautels_embeds_bb_{c_type}/{idx_}.p'), 'wb') as a:
                    pickle.dump(emb, a)


    elif args.mode == '4_train_model':
        for s in [0.2, 0.1, 0.05, 0.02, 0.01, 0.005]: # 0.2, 0.1, 0.05, 0.02, 0.01, 
            print("Start Training...")
            start_time = time.time()
            print("Train/Test Split: {}".format(s))
            if args.is_onehot:
                print("Training one-hot embedding based model")
                train_model_onehot(args, device_num = args.device_num, num_epochs = 100, batch_size = 32, 
                        exec_type = 'train', lr=0.01, traintest_split = s)
            else:
                print("Training regular embedding based model")
                train_model(args, device_num = args.device_num, num_epochs = 100, batch_size = 32, 
                        exec_type = 'train', lr=0.01, ridge=True, traintest_split = s, embed_type=args.embed_type)
            
            end_time = time.time()
            print("Done Training! Took {:.1f} Seconds.".format(end_time-start_time))
            print()

    elif args.mode == '5_generate_features':
        generate_features(args)

    elif args.mode == 'generate_h5_from_tens':
        generate_h5_from_tens(args)

    elif args.mode == '7_calc_topk_overlap':
        calc_topk_overlap(args, traintest_split = 0.02, r1 = 0.1, r2 = 0.1)

    else:
        assert False


from itertools import combinations
import os, sys
import pickle
from tqdm import tqdm
import argparse
import random
import numpy as np
from matplotlib import pyplot as plt
import time

import torch
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence
from torch import nn
from torch import optim as opt
import torch.nn.functional as F
from torch.optim.lr_scheduler import MultiStepLR
from typing import Callable, NamedTuple

from abmap.model import AbMAPAttn, MultiTaskLossWrapper
from abmap.utils import evaluate_spearman
from abmap.dataloader import create_dataloader_sabdab, create_dataloader_libra
sys.path.append('..') # for access to model.py when calling torch.load()


type_to_dim = {'beplerberger':6165, 'esm1b':1280, 'tape':768, 'dscript':100}


class TrainArguments(NamedTuple):
    cmd: str
    device_num: int
    log_file: str
    exec_type: str
    num_epochs: int
    embed_dir: str
    batch_size: int
    print_every: int
    ckpt_every: int
    lr: float
    gamma: float
    emb_type: str
    model_loadpath: str
    model_savepath: str
    mut_type: str
    chain_type: str
    region: str
    lambda1: float
    init_alpha: float
    update_alpha: bool


def add_args(parser):
    parser.add_argument('--device_num', type=int, default=0,
                        help='Indicate which GPU device to train the model on')

    parser.add_argument('--log_file', type=str,
                        help='Name of the log file with plot values.')

    parser.add_argument('--exec_type', type=str, default='train',
                        choices=['train', 'evaluate'],
                        help='Execution Mode.')

    parser.add_argument('--num_epochs', type=int, default=1,
                        help='Number of epochs for training.')

    parser.add_argument('--embed_dir', type=str, required=True,
                        help='Name of the folder that contains the embeddings')

    parser.add_argument('--batch_size', type=int, default=32,
                        help='Input batch size for the model.')

    parser.add_argument('--print_every', type=int, default=100,
                        help='Print every ... batch')

    parser.add_argument('--ckpt_every', type=int, default=5,
                        help='How often (in epochs) to save the model checkpoint.')

    parser.add_argument('--lr', type=float, default=0.01,
                        help='Model learning rate')

    parser.add_argument('--gamma', type=float, default=0.1,
                        help='Decay rate for the lr scheduler.')

    parser.add_argument('--emb_type', type=str, default='beplerberger',
                        help='Embedding Type.')

    parser.add_argument('--model_loadpath', type=str, default='None',
                        help='Path to an already trained model.')

    parser.add_argument('--model_savepath', type=str, default='',
                        help='Path to save the trained model.')

    parser.add_argument('--mut_type', type=str, default='cat2',
                        choices=['cat1', 'cat2'],
                        help='mutation type for random CDR mutations.')

    parser.add_argument('--chain_type', type=str, default='H',
                        choices=['H', 'L'],
                        help='Which chain in the Ab sequence to use.')

    parser.add_argument('--region', type=str, default='whole',
                        help='Which region on the sequence you would like to train on')

    parser.add_argument('--lambda1', type=float, default=0.0005,
                        help='Weight on the regularization loss')

    parser.add_argument('--init_alpha', type=float, default = 0.0,
                        help='Weight on the mse losses. 0 means an equal weight.')

    parser.add_argument('--update_alpha', action="store_true", dest='update_alpha',
                        help='The alpha value for the mse losses will be updated iteratively.')

    return parser


def train_model(device_num, log_file, exec_type, num_epochs, model_loadpath, model_savepath, region,
                embed_dir, batch_size, chain_type, init_alpha, update_alpha, lambda1, print_every=100, 
                lr=0.01, emb_type='beplerberger', ckpt_every=10, mut_type='cat2', gamma=0.5, **kwargs):
    start_time = time.time()

    if log_file == None:
        print("Your log will not be saved...")

    device = torch.device("cuda:{}".format(device_num) if torch.cuda.is_available() else "cpu")

    if embed_dir == "":
        print("You must provide the name of the directory for input embeddings")
        raise ValueError

    print("Loading data....")
    train_loader_sab, val_loader_sab, test_loader_sab = create_dataloader_sabdab(data_dir = embed_dir, 
                     batch_size=batch_size, emb_type=emb_type, mut_type=mut_type, chain_type=chain_type, region=region)
    train_loader_lib, val_loader_lib, test_loader_lib = create_dataloader_libra(data_dir = embed_dir, 
                     batch_size=batch_size, emb_type=emb_type, mut_type=mut_type, chain_type=chain_type)

    print("done loading!")
    train_size = len(train_loader_sab.dataset)

    # load model/operators
    if emb_type == 'beplerberger':
        model = AbMAPAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
    elif emb_type == 'protbert':
        model = AbMAPAttn(embed_dim=1024, mid_dim2=512, mid_dim3=256, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
    elif emb_type == 'esm1b' or emb_type == 'esm2':
        model = AbMAPAttn(embed_dim=1280, mid_dim2=512, mid_dim3=256, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
    elif emb_type == 'tape':
        model = AbMAPAttn(embed_dim=768, mid_dim2=256, mid_dim3=128, 
                                     proj_dim=60, num_enc_layers=1, num_heads=8).to(device)
    else:
        raise ValueError(f"The Embed Type {emb_type} Doesn't Exist!")

    loss_fn = nn.MSELoss()
    loss_wrapper = MultiTaskLossWrapper(2, init_alpha, update_alpha).to(device)
    # loss_opt = opt.SGD(loss_wrapper.parameters(), lr=lr)
    
    if model_loadpath == 'None':
        prev_epochs = 0
        optimizer = opt.SGD([{'params': model.parameters()},
                             {'params': loss_wrapper.parameters(), 'lr': 1e-3},], lr=lr)
        scheduler = MultiStepLR(optimizer, milestones=[40, 45], gamma=gamma)
    else:
        checkpoint = torch.load(model_loadpath, map_location=torch.device("cuda:{}".format(device_num)))
        prev_epochs = checkpoint['epoch']
        model.load_state_dict(checkpoint['model_state_dict'])
        print("Loaded a saved model from provided path!")
        # optimizer.load_state_dict(checkpoint['optimizer_state_dict'])

    if model_savepath == '':
        print("Not saving the model checkpoints.")

    if exec_type == 'train':

        # writer = SummaryWriter(log_dir='../../runs')
        stept, stepv = 0, 0
        train_spearmans_s, train_spearmans_f, val_spearmans_s, val_spearmans_f = [], [], [], []
        train_losses, val_losses = [], []
        for i in range(num_epochs):

            total_loss, total_mse_loss, total_reg_loss = 0, 0, 0
            total_mse_loss_s, total_mse_loss_f = 0, 0

            ########################################################################
            # TRAINING
            ########################################################################
            model.train()
            loss_wrapper.train()

            train_preds_s, train_targs_s, train_preds_f, train_targs_f = [], [], [], []

            assert len(train_loader_sab) == len(train_loader_lib)
            
            batch_count = 0
            for k, (batch_s, batch_f) in tqdm(enumerate(zip(train_loader_sab, train_loader_lib)), total=len(train_loader_sab)):
                batch_count += 1

                x1_s, x2_s, tmscore, x1_s_padding_mask, x2_s_padding_mask = batch_s
                x1_f, x2_f, func_sc, x1_f_padding_mask, x2_f_padding_mask = batch_f

                optimizer.zero_grad()
                # loss_opt.zero_grad()
                
                x1_s, x2_s = x1_s.cuda(device), x2_s.cuda(device)
                x1_f, x2_f = x1_f.cuda(device), x2_f.cuda(device)
                x1_s_padding_mask, x2_s_padding_mask = x1_s_padding_mask.to(device), x2_s_padding_mask.to(device)
                x1_f_padding_mask, x2_f_padding_mask = x1_f_padding_mask.to(device), x2_f_padding_mask.to(device)

                # feed the input data into the model
                try:
                    pred_s, e1_s, e2_s = model(x1_s, x2_s, x1_s_padding_mask, x2_s_padding_mask, 0)
                    pred_f, e1_f, e2_f = model(x1_f, x2_f, x1_f_padding_mask, x2_f_padding_mask, 1)
                    # unit_norm = F.normalize(torch.ones(e1_s.shape)).cuda(device)
                except:
                    print(x1_s.shape, x2_s.shape)
                    print(x1_f.shape, x2_f.shape)
                    print('input min maxes', torch.max(x1_s), torch.min(x1_s), torch.max(x1_f), torch.min(x1_f),
                          torch.max(x2_s), torch.min(x2_s), torch.max(x2_f), torch.min(x2_f))
                    raise ValueError

                # compute loss and back propagate
                reg_term = torch.zeros(e1_s.shape).cuda(device)
                for em in [e1_s, e2_s, e1_f, e2_f]:
                    em_sq = em ** 2
                    reg_term += em_sq*torch.log(em_sq + 1e-7)
                reg_loss = lambda1*torch.sum(reg_term)

                mse_loss = loss_wrapper(pred_s, tmscore.cuda(device), pred_f, func_sc.cuda(device))

                total_reg_loss += reg_loss.item()
                total_mse_loss += mse_loss.item()
                total_mse_loss_s += loss_wrapper.loss_s.item()
                total_mse_loss_f += loss_wrapper.loss_f.item()
                loss = mse_loss + reg_loss
                
                total_loss += loss.item()
                loss.backward()
                optimizer.step()
                # loss_opt.step()

                # for evaluation
                train_preds_s.append(pred_s)
                train_preds_f.append(pred_f)
                train_targs_s.append(tmscore.cuda(device))
                train_targs_f.append(func_sc.cuda(device))

                # print intermediate loss values
                if k % print_every == 0:
                    print("TRAIN Epoch {}, Batch {}, MSE Loss S: {}, MSE Loss F: {}, Reg Loss: {}, Alpha: {}".format(prev_epochs+i+1, 
                                            k, total_mse_loss_s/(k+1), total_mse_loss_f/(k+1), total_reg_loss/(k+1), loss_wrapper.alpha.item()))

            train_losses.append((total_loss/batch_count, total_mse_loss/batch_count, total_reg_loss/batch_count))

            # computing spearman rank scores
            train_preds_s, train_targs_s = torch.cat(train_preds_s), torch.cat(train_targs_s)
            train_preds_f, train_targs_f = torch.cat(train_preds_f), torch.cat(train_targs_f)
            spear_s = evaluate_spearman(train_preds_s, train_targs_s)
            spear_f = evaluate_spearman(train_preds_f, train_targs_f)
            train_spearmans_s.append(spear_s)
            train_spearmans_f.append(spear_f)

            # print learned loss ratio and Spearman ranks scores every epoch
            # print("Loss weights:", torch.exp(-1*loss_wrapper.weights))
            print("Epoch {} [Training] Model Evaluation <Sabdab> (Spearman): {}".format(prev_epochs+i+1, spear_s))
            print("Epoch {} [Training] Model Evaluation <LibraSeq> (Spearman): {}".format(prev_epochs+i+1, spear_f))
            # print("TRAIN Epoch {}, MSE Loss: {}, Reg Loss: {}, Alpha: {}".format(prev_epochs+i+1, 
            #                                  total_mse_loss/batch_count, total_reg_loss/batch_count, alpha_print))
            print('----------------------------------------------------------')


            ########################################################################
            # VALIDATION SET
            ########################################################################
            total_loss, total_reg_loss, total_mse_loss = 0, 0, 0
            model.eval()
            loss_wrapper.eval()
            val_preds_s, val_targs_s, val_preds_f, val_targs_f = [], [], [], []

            batch_count = 0
            for k, (batch_s, batch_f) in tqdm(enumerate(zip(val_loader_sab, val_loader_lib)), total=len(val_loader_sab)):
                batch_count += 1
                x1_s, x2_s, tmscore, x1_s_padding_mask, x2_s_padding_mask = batch_s
                x1_f, x2_f, func_sc, x1_f_padding_mask, x2_f_padding_mask = batch_f

                # feed the input with no gradient calculation
                with torch.no_grad():
                    x1_s, x2_s = x1_s.cuda(device), x2_s.cuda(device)
                    x1_f, x2_f = x1_f.cuda(device), x2_f.cuda(device)

                    x1_s_padding_mask, x2_s_padding_mask = x1_s_padding_mask.to(device), x2_s_padding_mask.to(device)
                    x1_f_padding_mask, x2_f_padding_mask = x1_f_padding_mask.to(device), x2_f_padding_mask.to(device)

                    pred_s, e1_s, e2_s = model(x1_s, x2_s, x1_s_padding_mask, x2_s_padding_mask, 0)
                    pred_f, e1_f, e2_f = model(x1_f, x2_f, x1_f_padding_mask, x2_f_padding_mask, 1)

                    # LOSS CALCULATIONS
                    reg_term = torch.zeros(e1_s.shape).cuda(device)
                    for em in [e1_s, e2_s, e1_f, e2_f]:
                        em_sq = em ** 2
                        reg_term += em_sq*torch.log(em_sq + 1e-7)
                    reg_loss = lambda1*torch.sum(reg_term)
                    # reg_loss = lambda1*torch.sum((e1_s - unit_norm)**2 + (e2_s - unit_norm)**2 + (e1_f - unit_norm)**2 + (e2_f - unit_norm)**2)

                    mse_loss = loss_wrapper(pred_s, tmscore.cuda(device), pred_f, func_sc.cuda(device))


                    total_reg_loss += reg_loss.item()
                    total_mse_loss += mse_loss.item()
                    loss = mse_loss + reg_loss

                total_loss += loss.item()

                # for evaluation
                val_preds_s.append(pred_s)
                val_preds_f.append(pred_f)
                val_targs_s.append(tmscore.cuda(device))
                val_targs_f.append(func_sc.cuda(device))

            val_losses.append((total_loss/batch_count, total_mse_loss/batch_count, total_reg_loss/batch_count))

            # computing spearman rank scores
            val_preds_s, val_targs_s = torch.cat(val_preds_s), torch.cat(val_targs_s)
            val_preds_f, val_targs_f = torch.cat(val_preds_f), torch.cat(val_targs_f)
            spear_s = evaluate_spearman(val_preds_s, val_targs_s)
            spear_f = evaluate_spearman(val_preds_f, val_targs_f)
            val_spearmans_s.append(spear_s)
            val_spearmans_f.append(spear_f)

            # print("Epoch {} Average Validation Loss. MSE: {}, Reg: {}".format(prev_epochs+i+1, total_mse_loss/batch_count, total_reg_loss/batch_count))
            # print("Loss weights:", torch.exp(-1*loss_wrapper.weights))
            print("VALIDATION Epoch {} Model Evaluation <Sabdab> (Spearman): {}".format(prev_epochs+i+1, spear_s))
            print("VALIDATION Epoch {} Model Evaluation <LibraSeq> (Spearman): {}".format(prev_epochs+i+1, spear_f))

            print('----------------------------------------------------------')

            # saving models:
            if (i+1)%ckpt_every == 0 and model_savepath != "":
                if not os.path.isdir(model_savepath):
                    os.mkdir(model_savepath)

                raw_model = model.module if hasattr(model, "module") else model
                model_sum = raw_model.state_dict()
                savepath = os.path.join(model_savepath, 'AbMAP_{}_{}_epoch{}.pt'.format(emb_type, chain_type, prev_epochs+i+1))
                torch.save(model_sum, savepath)

            # learning rate schedule
            scheduler.step()

    # test set
    print("Evaluating on Test Set...")
    model.eval()
    logs = [dict(), dict()]
    for k, test_loader in enumerate([test_loader_sab, test_loader_lib]):
        test_preds, test_targs = [], []
        cos = nn.CosineSimilarity(dim=1)
        tms_list, cos_list, points_list, pred_list = [], [], [], []

        for batch in tqdm(test_loader):
            x1_data, x2_data, tmscore, x1_padding_mask, x2_padding_mask = batch

            with torch.no_grad():
                x1 = x1_data.cuda(device)
                x2 = x2_data.cuda(device)

                x1_padding_mask, x2_padding_mask = x1_padding_mask.to(device), x2_padding_mask.to(device)

                pred, _, _ = model(x1, x2, x1_padding_mask, x2_padding_mask, k) # to criss-cross structure and function, use: (k+1)%2

                model = model.module if hasattr(model, "module") else model

                x1_mean = model.mean_over_seq(x1, x1_padding_mask)
                x2_mean = model.mean_over_seq(x2, x2_padding_mask)

                # x1_mean, x2_mean = model(x1, x2, x1_padding_mask, x2_padding_mask, task=k, task_specific=True)

                # for evaluation
                test_preds.append(pred)
                test_targs.append(tmscore.cuda(device))

            pred_list += pred.tolist()

            # x1_mean = torch.mean(x1_p, dim=1)
            # x2_mean = torch.mean(x2_p, dim=1)

            cos_sim = cos(x1_mean, x2_mean)
            cos_temp = cos_sim.tolist()
            cos_list += cos_temp

            tms_temp = tmscore.tolist()
            tms_list += tms_temp

            for i in range(len(tms_temp)):
                points_list.append((tms_temp[i], cos_temp[i]))

        test_preds = torch.cat(test_preds)
        test_targs = torch.cat(test_targs)
        spear = evaluate_spearman(test_preds, test_targs)
        print("Task {}, Test set spearman rank: {}".format(k, spear))


        sorted_pairs = sorted(points_list, key=lambda x:x[0], reverse=False)

        split_arrays = np.array_split(sorted_pairs, 10)

        means = []
        for bucket in split_arrays:
            temp = []
            for tms, dot in bucket:
                temp.append(dot)
            means.append(np.mean(temp))

        logs[k]["bucket"] = means

        logs[k]["scatter_proj"] = [tms_list, cos_list]

        logs[k]["scatter_pred"] = [tms_list, pred_list]

        if exec_type == 'train':
        # Creating Loss and Spearman Plots
            if k == 0:
                plots_dict = {'train_losses': train_losses, 'val_losses': val_losses,
                              'train_spearmans': train_spearmans_s, 'val_spearmans':val_spearmans_s}
            else:
                plots_dict = {'train_losses': train_losses, 'val_losses': val_losses,
                              'train_spearmans': train_spearmans_f, 'val_spearmans':val_spearmans_f}

            logs[k]["train_losses"] = plots_dict['train_losses']
            logs[k]["val_losses"] = plots_dict['val_losses']
            logs[k]["train_spearmans"] = plots_dict['train_spearmans']
            logs[k]["val_spearmans"] = plots_dict['val_spearmans']

    if log_file != None:
        with open(log_file, "wb") as f:
            pickle.dump(logs, f)

    end_time = time.time()
    print("It took {:.2f} minutes to train!".format((end_time-start_time)/60))


def main(args):

    print(args)
    train_model(**vars(args))


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    add_args(parser)

    args = parser.parse_args()

    if torch.cuda.is_available():
        print("Using the GPU!")
    else:
        print("WARNING: Could not find GPU! Using CPU only")

    torch.multiprocessing.set_start_method('spawn')
    main(args)
    print("DONE")


from tqdm import tqdm
import os
import string
import pickle
import sys
import pandas as pd
import multiprocessing as mp
from itertools import permutations, combinations
import random
import glob
import numpy as np
import csv

from abmap.abmap_augment import ProteinEmbedding
from abmap.utils import find_sequence
from abmap.plm_embed import reload_models_to_device

import psico.fullinit
from pymol import cmd

def init(dev_num, c_type, region, which_set):
    global chain_type, prot_region
    chain_type = c_type
    prot_region = region

    global data_path, id_pairs_path
    data_path = '../data/raw/sabdab/sabdab_dataset'
    id_pairs_path = '../data/processed/sabdab/all_pairs_{}.p'.format(which_set)


def work_parallel(idx_pair):
    idx, pdb_id1, pdb_id2 = idx_pair

    cmd.delete('all')

    p1_path = os.path.join(data_path, pdb_id1, 'structure', '{}.pdb'.format(pdb_id1))
    p2_path = os.path.join(data_path, pdb_id2, 'structure', '{}.pdb'.format(pdb_id2))

    cmd.load(p1_path, "999C")
    cmd.load(p2_path, "999D")

    cmd.split_chains('999C')
    cmd.split_chains('999D')


    p1_seq_H, p1_seq_L = find_sequence(pdb_id=pdb_id1)
    p2_seq_H, p2_seq_L = find_sequence(pdb_id=pdb_id2)

    if chain_type == 'H':
        p1_seq, p2_seq = p1_seq_H, p2_seq_H
    else:
        p1_seq, p2_seq = p1_seq_L, p2_seq_L

    fold = prot_region if chain_type == 'H' else str(5 + int(prot_region))
    start1, end1, start2, end2 = 0, 0, 0, 0
    if prot_region != 'whole':
        try:
            p1 = ProteinEmbedding(p1_seq, chain_type=chain_type, dev=idx, fold=fold)
            p1.create_cdr_mask()
            p2 = ProteinEmbedding(p2_seq, chain_type=chain_type, dev=idx, fold=fold)
            p2.create_cdr_mask()
        except:
            return (idx, -1, -1)

        p1_mask = p1.cdr_mask.tolist()
        a = list(map(int, p1_mask))
        t = list(map(str, a))
        p1_mask_str = "".join(t)

        p2_mask = p2.cdr_mask.tolist()
        a = list(map(int, p2_mask))
        t = list(map(str, a))
        p2_mask_str = "".join(t)

        if prot_region not in p1_mask_str or prot_region not in p2_mask_str:
            return (idx, -1, -1)

        start1, end1 = p1_mask_str.index(prot_region)+1, p1_mask_str.rindex(prot_region)+1
        start2, end2 = p2_mask_str.index(prot_region)+1, p2_mask_str.rindex(prot_region)+1

        del p1, p2, p1_mask, p2_mask, p1_mask_str, p2_mask_str # prevents multiprocessing from slowing down!

        # try:
        #     start1, end1 = p1_mask_str.index(prot_region)+1, p1_mask_str.rindex(prot_region)+1
        # except:
        #     print("p1 mask:", p1_mask_str)
        #     raise ValueError
        # try:
        #     start2, end2 = p2_mask_str.index(prot_region)+1, p2_mask_str.rindex(prot_region)+1
        # except:
        #     print("p2 mask:", p2_mask_str)
        #     raise ValueError

        # print("Indices for Chain {} Region {}, Against {} :".format(chain_type, prot_region, against), start1, end1, start2, end2)


    # find the chain letters
    chain_letters = []
    for pdb_id in [pdb_id1, pdb_id2]:
        h = glob.glob(os.path.join(data_path, pdb_id, 'sequence', '{}_*_V{}.fa'.format(pdb_id, chain_type)))
        h1 = [a for a in h if "{}_V{}.fa".format(chain_type, chain_type) in a]
        if len(h1)> 0:
           vhfile = h1[0]
        else:
           vhfile = h[0]
        chain_letter = vhfile[-7]
        chain_letters.append(chain_letter)

    prot1 = '999C_{}'.format(chain_letters[0])
    prot2 = '999D_{}'.format(chain_letters[1])


    if prot_region == 'whole':
        try:
            rmsd = cmd.super(prot1+'////CA', prot2+'////CA')[0]
        except:
            print("{},{}-{} RMSD for region {} didn't work...".format(pdb_id1, pdb_id2, chain_type, prot_region))
            rmsd = -1
        try:
            tm_score = cmd.tmalign(prot1+'////CA', prot2+'////CA')
        except:
            print("{},{}-{} TM Score for region {} didn't work...".format(pdb_id1, pdb_id2, chain_type, prot_region))
            tm_score = -1

    else:
        try:
            rmsd = cmd.super(prot1+'///{}-{}/CA'.format(start1, end1), prot2+'///{}-{}/CA'.format(start2, end2))[0]
        except:
            print("{},{}-{} RMSD for region {} didn't work...".format(pdb_id1, pdb_id2, chain_type, prot_region))
            rmsd = -1
        try:
            tm_score = cmd.tmalign(prot1+'///{}-{}/CA'.format(start1, end1), prot2+'///{}-{}/CA'.format(start2, end2))
        except:
            print("{},{}-{} TM Score for region {} didn't work...".format(pdb_id1, pdb_id2, chain_type, prot_region))
            tm_score = -1

    return (idx, tm_score, rmsd)


def get_valid_ids():
    with open('../data/processed/sabdab/valid_ids_all.txt', 'w') as f:
        valid_list = []
        path = '../data/raw/sabdab/sabdab_dataset'
        pdb_ids = os.listdir(path)
        for pdb_id in tqdm(pdb_ids):
            try:
                vhs, vls = find_sequence(pdb_id = pdb_id)
                prot_embed_h = ProteinEmbedding(vhs, 'H')
                prot_embed_l = ProteinEmbedding(vls, 'L')
                prot_embed_h.create_cdr_mask()
                prot_embed_l.create_cdr_mask()
                valid_list.append(pdb_id)
            except:
                pass
        f.write('\n'.join(valid_list))


def generate_id_pairs(args):
    valid_ids_path = '../data/processed/sabdab/valid_ids_{}.txt'.format(args.set)

    # saving perms
    with open(valid_ids_path, 'r') as f:
        valid_ids = f.read().splitlines()
    perms = list(combinations(valid_ids, 2))
    random.shuffle(perms)
    
    with open('../data/processed/sabdab/all_pairs_{}.p'.format(args.set), 'wb') as f:
        pickle.dump(perms, f)

    print('DONE!')
    return


def save_sabdab_cdrs(args):

    pdb_ids_path = "../data/processed/sabdab/valid_ids_all.txt"
    with open(pdb_ids_path, 'r') as f:
        pdb_ids = f.read().splitlines()

    results_dict_H, results_dict_L = dict(), dict()
    for pdb_id in tqdm(pdb_ids):
        seq_h, seq_l = find_sequence(pdb_id=pdb_id)

        prot_h = ProteinEmbedding(seq_h, 'H')
        prot_h.create_cdr_mask()
        prot_l = ProteinEmbedding(seq_l, 'L')
        prot_l.create_cdr_mask()

        cdrs_H = ["", "", ""]
        for i in range(len(prot_h.cdr_mask)):
            for q in (1, 2, 3):
                if prot_h.cdr_mask[i] == q:
                    cdrs_H[q-1] += seq_h[i]

        cdrs_L = ["", "", ""]
        for i in range(len(prot_l.cdr_mask)):
            for q in (1, 2, 3):
                if prot_l.cdr_mask[i] == q:
                    cdrs_L[q-1] += seq_l[i]

        results_dict_H[pdb_id] = cdrs_H
        results_dict_L[pdb_id] = cdrs_L

    with open('../data/processed/sabdab/sabdab_cdrH_strs.p', 'wb') as p:
        pickle.dump(results_dict_H, p)
    with open('../data/processed/sabdab/sabdab_cdrL_strs.p', 'wb') as p:
        pickle.dump(results_dict_L, p)


def main_sabdab(args, orig_embed = False):

    pdb_ids = os.listdir('../data/raw/sabdab/sabdab_dataset')[:10]

    print(len(pdb_ids))

    out_folder = "../data/processed/sabdab/cdrembed_maskaug4"
    # out_folder = "/net/scratch3/scratch3-3/chihoim/ablm/data/processed/sabdab/cdrembed_maskaug4"
    if not os.path.isdir(out_folder):
        os.mkdir(out_folder)

    embed_type = 'beplerberger'
    # embed_type = 'protbert'
    # embed_type = 'esm1b'
    # embed_type = 'tape'

    reload_models_to_device(args.device_num)

    k=100
    for c_type in ['H', 'L']:
        for pdb_id in tqdm(pdb_ids):

            file_name = '{}_{}_{}_k{}.p'.format(pdb_id, 'cat2', c_type, k)
            if os.path.exists(os.path.join(out_folder, embed_type, file_name)):
                continue


            seq_h, seq_l = find_sequence(pdb_id = pdb_id)
            seq = seq_h if c_type == 'H' else seq_l
            prot_embed = ProteinEmbedding(seq, c_type, fold='x')

            out_path = os.path.join(out_folder, embed_type)
            if not os.path.isdir(out_path):
                os.mkdir(out_path)


            try:
                cdr_embed = prot_embed.create_cdr_specific_embedding(embed_type, k=k, seperator=False, mask=True)
            except:
                print("processing pdb id {} didn't work...".format(pdb_id))
                # continue

            file_name = '{}_{}_{}_k{}.p'.format(pdb_id, 'cat2', c_type, k)
            with open(os.path.join(out_path, file_name), 'wb') as fh:
                print("Saving", pdb_id)
                pickle.dump(cdr_embed, fh)



def make_sabdab_features(args):
    reload_models_to_device(args.device_num)

    device = torch.device(base_config.device if torch.cuda.is_available() else "cpu")
    
    chain_type = args.chain_type
    out_dir = "../data/processed/sabdab/{}_features".format(args.set)
    embed_path = "../data/processed/sabdab/cdrembed_maskaug4/beplerberger"
    pdb_ids_path = "..data/processed/sabdab/valid_ids_{}.txt".format(args.set)
    with open(pdb_ids_path, 'r') as f:
        pdb_ids = f.read().splitlines()

    # Load our pretrained AbNet Model
    from model import AntibodyNetMultiAttn
    pretrained = AntibodyNetMultiAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
    pretrained_path = "../model_ckpts/091522_bb_newnum_{}/beplerberger_epoch50.pt".format(chain_type)
    checkpoint = torch.load(pretrained_path, map_location=device)
    pretrained.load_state_dict(checkpoint['model_state_dict'])
    pretrained.eval()
    print("Loaded the pre-trained model!")

    k = 100
    out_path = os.path.join(out_dir, 'beplerberger', chain_type)
    for pdb_id in tqdm(pdb_ids):
        file_name = 'sabdab_{}_{}_{}_k{}.p'.format(pdb_id, 'cat2', chain_type, k)
        with open(os.path.join(embed_path, file_name), 'rb') as f:
            prot_emb = pickle.load(f)

        cdr_embed = torch.unsqueeze(prot_emb, dim=0) # create the batch dimension (singleton)
        cdr_embed = cdr_embed.to(device)

        # create and save STRUCTURE specific feature:
        with torch.no_grad():
            feat_task, _ = pretrained(cdr_embed, cdr_embed, None, None, 0, task_specific=True)
        assert feat_task.shape[-1] == 512
        with open(os.path.join(out_path, "{}_struc_{}.p".format(pdb_id, chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_task), p)

        # create and save FUNCTION specific feature:
        with torch.no_grad():
            feat_task, _ = pretrained(cdr_embed, cdr_embed, None, None, 1, task_specific=True)
        assert feat_task.shape[-1] == 512
        with open(os.path.join(out_path, "{}_func_{}.p".format(pdb_id, chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_task), p)

        # create and save INTERMEDIATE n' x k feature, before Transformer:
        with torch.no_grad():
            feat_interm, _ = pretrained(cdr_embed, cdr_embed, None, None, None, return2=True)
        assert feat_interm.shape[-1] == 256
        with open(os.path.join(out_path, "{}_interm_{}.p".format(pdb_id, chain_type)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_interm), p)

    return




if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", help="which code path to run. see main(..) for details", type=str)
    parser.add_argument("--device_num", help="GPU device for computation.", type=int, default=0)
    parser.add_argument("--num_processes", help="number of processes for multiprocessing", type=int, default=1)
    parser.add_argument("--chunk_size", help="chunk size for each processing step", type=int, default=1)
    parser.add_argument("--region", help="region of the protein sequence that we're processing", type=str)
    parser.add_argument("--chain_type", help="Which chain to process: H or L", type=str, choices = ["H", "L"])
    parser.add_argument("--set", help="Creating set1 or set2 dataset", type=str, choices = ["set1", "set2"])
    parser.add_argument("--extra", help="put this as the LAST option and arbitrary space-separated key=val pairs after that", type=str, nargs='*')

    args = parser.parse_args()
    assert args.mode is not None
    
    extra_args = dict([a.split("=") for a in args.extra]) if args.extra else {}
    args.extra = extra_args
    
    pd.set_option('use_inf_as_na', True)

    
    if args.mode == '1_generate_id_pairs':
        generate_id_pairs(args)
    elif args.mode == '2_calculate_pair_scores':
        sample_n = 110000

        # load the pairs:
        all_pairs_path = '../data/processed/sabdab/all_pairs_{}.p'.format(args.set)
        with open(all_pairs_path, 'rb') as f:
            all_pairs = pickle.load(f)
        n_pairs = all_pairs[:sample_n]
        n_pairs = [(i, id1, id2) for i, (id1, id2) in enumerate(n_pairs)]

        # perform multiprocessing
        import multiprocessing as mp
        from multiprocessing import get_context

        tm_scores = set()
        with get_context('spawn').Pool(processes=args.num_processes, initializer=init, 
                                       initargs=[args.device_num, args.chain_type, args.region, args.set]) as p:
            for result in tqdm(p.imap_unordered(work_parallel, n_pairs, chunksize=args.chunk_size), total=len(n_pairs)):
                tm_scores.add(result)

                # to save intermediate results. Comment out as necessary
                if len(tm_scores)%10000 == 0:
                    with open('../data/processed/sabdab/CDR{}{}_pairs_{}.p'.format(args.chain_type, args.region, len(tm_scores)), 'wb') as f:
                        pickle.dump(tm_scores, f)


        tm_scores = list(tm_scores)
        tm_scores.sort(key=lambda x: x[0])

        # save the results into a csv file
        # save_count, num_pairs = 0, 100000
        with open('../data/processed/sabdab/{}_CDR{}_{}_pair_scores_100k.csv'.format(args.set, args.region, args.chain_type), 'w', encoding='UTF8') as f:
            writer = csv.writer(f)
            writer.writerow(['pdb_id1', 'pdb_id2', 'tm_score', 'rmsd'])
            for i in range(len(tm_scores)):
                if tm_scores[i][1] > 0 and tm_scores[i][2] >= 0:
                    writer.writerow([n_pairs[i][1], n_pairs[i][2], tm_scores[i][1], tm_scores[i][2]])
                    # save_count += 1

                    # if save_count == num_pairs:
                    #     break

    elif args.mode == '3_generate_plm_embeddings':
        main_sabdab(args)

    else:
        assert False


import os
import argparse
from tqdm import tqdm
import glob
import time
import numpy as np
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt
import pickle
import pandas as pd

import sys
# sys.path.append('..')

from ..base_config import *
# import base_config
from ..abmap.utils import get_boolean_mask, find_sequence
from mutate import generate_mutation
from embed import embed_sequence, reload_models_to_device #, load_model
from model import MultiTaskLossWrapper

import torch

print("all imports were successful!")
assert False



def create_covabdab_embeddings(args):
    reload_models_to_device(args.device_num)

    df = pd.read_csv("/data/cb/rsingh/work/antibody/ci_data/raw/covabdab/neutralize_sarscov2.csv")
    device = torch.device(base_config.device if torch.cuda.is_available() else "cpu")
    out_dir = "/data/cb/rsingh/work/antibody/ci_data/processed/covabdab/"
    chain_type = 'H'

    # load valid ids:
    with open('/data/cb/rsingh/work/antibody/ci_data/processed/covabdab/valid_ids.txt', 'r') as f:
        valid_ids_list = f.read().splitlines()

    k=100
    invalids = []

    for index, row in tqdm(df.iterrows(), total=df.shape[0]):
        # if index < 1710:
        #     continue

        seq_id = row['Name']
        if seq_id not in valid_ids_list:
            continue

        seq = row['VH or VHH'] if chain_type == 'H' else row['VL']

        if ' ' in seq:
            invalids.append(seq_id)
            continue

        prot_embed = ProteinEmbedding(seq, chain_type, dev=args.device_num)
        try:
            prot_embed.create_cdr_mask()
            # print(prot_embed.cdr_mask)
            # print("HAW"+prot_embed.seq+"HEE")
        except:
            invalids.append(seq_id)
            continue

        for embed_type in ['beplerberger']:
            out_path = os.path.join(out_dir, 'cdrembed_maskaug4', embed_type)
            prot_embed.embed_seq(embed_type=embed_type)
            kmut_matr_h = prot_embed.create_kmut_matrix(num_muts=k, embed_type=embed_type)
            cdr_embed = prot_embed.create_cdr_embedding(kmut_matr_h, sep=False, mask=True)

            # save the embedding first:
            with open(os.path.join(out_path, "{}_cat2_{}_k{}.p".format(seq_id, chain_type, k)), 'wb') as fc:
                pickle.dump(cdr_embed, fc)

    print(invalids)



def create_covabdab_features(args):
    reload_models_to_device(args.device_num)

    df = pd.read_csv("/data/cb/rsingh/work/antibody/ci_data/raw/covabdab/neutralize_sarscov2.csv")
    device = torch.device(base_config.device if torch.cuda.is_available() else "cpu")
    covabdab_dir = "/data/cb/rsingh/work/antibody/ci_data/processed/covabdab"
    embed_dir = "/data/cb/rsingh/work/antibody/ci_data/processed/covabdab/cdrembed_maskaug4/{}".format(args.embed_type)
    out_dir = "/data/cb/rsingh/work/antibody/ci_data/processed/covabdab/covabdab_features_120122/{}".format(args.embed_type)

    # Load our pretrained AbNet Model
    from model import AntibodyNetMultiAttn
    model = AntibodyNetMultiAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512,
                                      proj_dim=252, num_enc_layers=1, num_heads=16).to(device)
    pretrained_path = "../model_ckpts/091522_bb_newnum_{}/beplerberger_epoch50.pt".format(args.chain_type)
    checkpoint = torch.load(pretrained_path, map_location=device)
    model.load_state_dict(checkpoint['model_state_dict'])
    model.eval()
    print("Loaded the pre-trained model!")

    k=100

    with open(os.path.join(covabdab_dir, 'invalids.txt'), 'r') as f:
        invalids_list = f.read().splitlines()

    for index, row in tqdm(df.iterrows(), total=df.shape[0]):
        seq_id, seq_h = row['Name'], row['VH or VHH']

        # if the sequence id is invalid, skip
        if seq_id in invalids_list:
            continue

        # prot_embed = ProteinEmbedding(seq_h, 'H', dev=1)
        # prot_embed.create_cdr_mask()

        # --------------------------------------
        # FOR PROTBERT ONLY:

        # embed_type = args.embed_type
        # prot = ProteinEmbedding(seq_h, 'H')
        # try:
        #     prot.embed_seq(embed_type = args.embed_type)
        # except:
        #     continue
        # with open(os.path.join(out_dir, embed_type, '{}_H_orig.p'.format(seq_id)), 'wb') as f:
        #     pickle.dump(prot.embedding, f)
        # continue
        # --------------------------------------

        # load the embedding:
        with open(os.path.join(embed_dir, "{}_cat2_H_k100.p".format(seq_id)), 'rb') as p:
            cdr_embed = pickle.load(p)

        cdr_embed = cdr_embed.to(device)
        cdr_embed = torch.unsqueeze(cdr_embed, dim=0) # create the batch dimension (singleton)

        # create and save STRUCTURE specific feature:
        with torch.no_grad():
            feat_task, _ = model(cdr_embed, cdr_embed, None, None, 0, task_specific=True)
        assert feat_task.shape[-1] == 512
        with open(os.path.join(out_dir, "{}_struc_H.p".format(seq_id)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_task), p)

        # create and save FUNCTION specific feature:
        with torch.no_grad():
            feat_task, _ = model(cdr_embed, cdr_embed, None, None, 1, task_specific=True)
        assert feat_task.shape[-1] == 512
        with open(os.path.join(out_dir, "{}_func_H.p".format(seq_id)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_task), p)

        # create and save INTERMEDIATE n' x k feature, before Transformer:
        with torch.no_grad():
            feat_interm, _ = model(cdr_embed, cdr_embed, None, None, None, return2=True)
        assert feat_interm.shape[-1] == 256
        with open(os.path.join(out_dir, "{}_interm_H.p".format(seq_id)), 'wb') as p:
            pickle.dump(torch.squeeze(feat_interm), p)


if __name__ == "__main__":
    print("running covabdab analysis!")


#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()
    
with open("abmap/version.py") as file:
    exec(file.read())

setup(
    name="abmap",
    version=version,
    description="AbMAP: Antibody Mutagenesis Augmented Processing",
    long_description=readme, 
    author="Chiho Im, Taylor Sorenson, Abhinav Gupta, Rohit Singh",
    author_email="chihoim@mit.edu",
    # url="http://dscript.csail.mit.edu",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "abmap = abmap.__main__:main",
        ],
    },
    include_package_data=True,
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "torch>=1.11",
        "biopython",
        "matplotlib",
        "seaborn",
        "tqdm",
        "scikit-learn",
        "h5py",
        "transformers",
        "positional_encodings",
        "dscript"
    ],
)



"""
Given fasta sequences and a pre-trained AbMAP model,
generate their AbMAP embeddings (fixed or variable).
"""

from __future__ import annotations
import argparse

from abmap.plm_embed import reload_models_to_device
from abmap.abmap_augment import ProteinEmbedding
from abmap.model import AbMAPAttn
from abmap.utils import parse
import sys
sys.path.append('../') # for access to model.py when calling torch.load()

from typing import Callable, NamedTuple
import torch
import pickle
from tqdm import tqdm
import os

class EmbedArguments(NamedTuple):
    cmd: str
    device: int
    chain_type: str
    pretrained_path: str
    input_path: str
    variable_length: bool
    task: int
    func: Callable[[EmbedArguments], None]


def add_args(parser):
    """
    Create parser for command line utility.
    :meta private:
    """
    parser.add_argument(
        "-d", "--device", type=int, default=-1, help="Compute device to use"
    )
    parser.add_argument(
        "--chain-type", dest="chain_type", choices=['H', 'L'],
        help="Chain type of the fasta sequences (H or L)", required=True
    )
    parser.add_argument(
        "--pretrained-path", dest="pretrained_path",
        help="path for the pre-trained AbMAP Model", required=True
    )
    parser.add_argument(
        "--input-file", dest="input_file",
        help="path for fasta file containing sequences to be embeded using the pre-trained AbMAP model", required=True
    )
    parser.add_argument(
        "--output-dir", dest="output_dir",
        help="path to save the output AbMAP embedding", required=True
    )
    parser.add_argument(
        "--plm-name", dest="plm_name",
        help="Name of the foundational PLM used when generating the augmented input embedding", default='beplerberger'
    )
    parser.add_argument(
        "--variable-length", action="store_true", dest='variable_length',
        help="Output variable length AbMAP Embeddings. (Default: Fixed Length)"
    )
    parser.add_argument(
        "--task", type=int, default='structure', choices=['structure', 'function'],
        help="Which task to be specified for generating the embedding (0: structure or 1: function). Default: 0"
    )
    return parser


def load_abmap(pretrained_path, plm_name, device=0):
    '''
    Load AbMAP weights and foundational model
    '''
    dev = torch.device(f'cuda:{device}' if torch.cuda.is_available() else "cpu")
    reload_models_to_device(device, plm_name)

    # load pre-trained model into device
    if plm_name == 'beplerberger':
        pretrained = AbMAPAttn(embed_dim=2200, mid_dim2=1024, mid_dim3=512, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(dev)
    if plm_name == 'protbert':
        pretrained = AbMAPAttn(embed_dim=1024, mid_dim2=512, mid_dim3=256, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(dev)
    if plm_name == 'esm1b':
        pretrained = AbMAPAttn(embed_dim=1280, mid_dim2=512, mid_dim3=256, 
                                     proj_dim=252, num_enc_layers=1, num_heads=16).to(dev)
    if plm_name == 'tape':
        pretrained = AbMAPAttn(embed_dim=768, mid_dim2=256, mid_dim3=128, 
                                     proj_dim=60, num_enc_layers=1, num_heads=8).to(dev)
    checkpoint = torch.load(pretrained_path, map_location=dev)
    if 'model_state_dict' in checkpoint:
        pretrained.load_state_dict(checkpoint['model_state_dict'])
    else:
        pretrained.load_state_dict(checkpoint)
    pretrained.eval()
    print("Loaded the Pre-trained Model!")
    return pretrained


def abmap_embed_batch(device, pretrained_path, input_path, variable_length, plm_name, task):
    '''
    Like abmap_embed, but works for a directory of inputs
    '''
    pretrained = load_abmap(pretrained_path, plm_name, device)
    dev = torch.device(f'cuda:{device}' if torch.cuda.is_available() else "cpu")

    # Get list of files or casts single file as a list
    if os.path.isdir(input_path): input_iter = os.listdir(input_path)
    elif os.path.isfile(input_path): input_iter = list(input_path)
    
    outputs = []
    for input in tqdm(input_iter):
        input_ = os.path.join(input_path, input)
        with open(input_, 'rb') as p:
            input_embed = pickle.load(p).to(dev)
        input_embed = torch.unsqueeze(input_embed, 0)
        try:
            assert len(input_embed.shape) == 3
        except:
            raise ValueError("input embedding should be of shape n'(CDR length) x d")

        # generate the abmap embedding
        with torch.no_grad():
            if variable_length:
                out_feature, _ = pretrained.embed(input_embed, task=task, embed_type='variable')
            else:
                out_feature, _ = pretrained.embed(input_embed, task=task, embed_type='fixed')
        out_feature = torch.squeeze(out_feature, 0)
        outputs.append(out_feature)
        
    return outputs

        
def abmap_embed(device, pretrained_path, chain_type, input_file, output_dir, variable_length, plm_name, task):

    """
    Given fasta sequences and a pre-trained AbMAP model,
    generate their AbMAP embeddings (fixed or variable).
    
    ***
    For Pre-Trained AbMAP models, the augmented embeddings have the following parameters:
        Foundational PLM: Bepler & Berger
        CDR-masks (4 extra dims)
        NO separators
        # mutations = 100
    """
    embed_type = 'beplerberger'
    task_ = 'structure' if task == 0 else 'function'
    output_type = 'variable' if variable_length else 'fixed'
    
    pretrained = load_abmap(pretrained_path, plm_name, device)
    dev = torch.device(f'cuda:{device}' if torch.cuda.is_available() else "cpu")

    # get the names and sequences of fasta files
    names, seqs = parse(input_file)
    
    # check if outputPath is an existing directory:
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    # generate the embeddings for fasta sequences:    
    for (name, seq) in tqdm(zip(names, seqs), total=len(names)):
        fname = os.path.join(output_dir, f'{name}_AbMAP.p')
        if not os.path.isfile(fname):

            prot = ProteinEmbedding(seq, chain_type, embed_device=f'cuda:{device}')
            z = prot.create_cdr_specific_embedding(embed_type)
            z = torch.unsqueeze(z, dim=0).to(dev)

            with torch.no_grad():
                output = pretrained.embed(z, task=task_, embed_type=output_type)
            output = torch.squeeze(output, dim=0)
            
            # save the abmap embedding
            with open(fname, 'wb') as f:
                pickle.dump(output.cpu(), f)
                
    return
                


def main(args):
    """
    Run embedding from arguments.
    :meta private:
    """
    device = args.device
    chain_type = args.chain_type
    pretrained_path = args.pretrained_path
    input_file = args.input_file
    output_dir = args.output_dir
    variable_length = args.variable_length
    plm_name = args.plm_name
    task = args.task
    abmap_embed(device, pretrained_path, chain_type, input_file, output_dir,
                 variable_length, plm_name, task)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    add_args(parser)
    main(parser.parse_args())
